from codemie_tools.base.models import ToolMetadata
from .models import AzureConfig

AZURE_TOOL = ToolMetadata(
    name="Azure",
    description="""
    Tool for interacting with Azure REST API.
    All credentials and authentication handling is done automatically by the tool.

    Capabilities:
    - Access to Azure Resource Manager REST API
    - Manage Azure resources (VMs, Storage, Networking, etc.)
    - Execute any Azure REST API operation
    - Support for multiple Azure services (management, graph, storage, database)
    - Automatic OAuth token management

    Usage:
    Provide HTTP method, full URL, and optional request arguments.
    The tool handles authentication and token refresh automatically.
    """.strip(),
    label="Azure",
    user_description="""
    Provides access to the Azure API, allowing for management and interaction with various Azure
    resources and services. This tool enables the AI assistant to perform a wide range of operations
    within the Azure cloud environment.

    Before using it, it is necessary to add a new integration for the tool by providing:
    1. Alias (A friendly name for the Azure account)
    2. Subscription ID
    3. Tenant ID
    4. Client ID
    5. Client Secret

    Usage Note:
    Use this tool when you need to manage Azure resources, deploy services, or retrieve information
    from your Azure environment.
    """.strip(),
    settings_config=True,
    config_class=AzureConfig
)



import logging
from typing import Type, Optional

import pymupdf
from langchain_core.tools import ToolException
from pydantic import BaseModel

from codemie_tools.base.codemie_tool import CodeMieTool
from .models import (
    ReportPortalConfig,
    GetExtendedLaunchDataInput,
    GetExtendedLaunchDataAsRawInput,
    GetLaunchDetailsInput,
    GetAllLaunchesInput,
    FindTestItemByIdInput,
    GetTestItemsForLaunchInput,
    GetLogsForTestItemInput,
    GetUserInformationInput,
    GetDashboardDataInput,
    UpdateTestItemInput
)
from .report_portal_client import ReportPortalClient
from .tools_vars import (
    GET_EXTENDED_LAUNCH_DATA_TOOL,
    GET_EXTENDED_LAUNCH_DATA_AS_RAW_TOOL,
    GET_LAUNCH_DETAILS_TOOL,
    GET_ALL_LAUNCHES_TOOL,
    FIND_TEST_ITEM_BY_ID_TOOL,
    GET_TEST_ITEMS_FOR_LAUNCH_TOOL,
    GET_LOGS_FOR_TEST_ITEM_TOOL,
    GET_USER_INFORMATION_TOOL,
    GET_DASHBOARD_DATA_TOOL,
    UPDATE_TEST_ITEM_TOOL
)

logger = logging.getLogger(__name__)


class BaseReportPortalTool(CodeMieTool):
    """Base class for Report Portal tools."""

    config: ReportPortalConfig
    client: Optional[ReportPortalClient] = None

    def __init__(self, config: ReportPortalConfig):
        """Initialize the tool with configuration."""
        super().__init__(config=config)
        if config.url and config.api_key and config.project:
            self.client = ReportPortalClient(
                endpoint=config.url,
                project=config.project,
                api_key=config.api_key
            )

    def _healthcheck(self):
        self.client.get_all_launches(page_number=1)


class GetExtendedLaunchDataTool(BaseReportPortalTool):
    """Tool to get extended launch data from Report Portal."""

    name: str = GET_EXTENDED_LAUNCH_DATA_TOOL.name
    description: str = GET_EXTENDED_LAUNCH_DATA_TOOL.description
    args_schema: Type[BaseModel] = GetExtendedLaunchDataInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, launch_id: str):
        """
        Get extended launch data.

        Args:
            launch_id: Launch ID of the launch to export

        Returns:
            str: Text content of the launch report

        Raises:
            ToolException: If export fails
        """
        try:
            format_type = 'html'
            response = self.client.export_specified_launch(launch_id, format_type)

            if response.headers['Content-Type'] in ['application/pdf', 'text/html']:
                with pymupdf.open(stream=response.content, filetype=format_type) as report:
                    text_content = ''
                    for page_num in range(len(report)):
                        page = report[page_num]
                        text_content += page.get_text()
                    return text_content
            else:
                logger.warning(f"Exported data for launch {launch_id} is in an unsupported format.")
                return None
        except Exception as e:
            logger.error(f"Error getting extended launch data: {str(e)}")
            raise ToolException(f"Error getting extended launch data: {str(e)}")


class GetExtendedLaunchDataAsRawTool(BaseReportPortalTool):
    """Tool to get extended launch data as raw from Report Portal."""

    name: str = GET_EXTENDED_LAUNCH_DATA_AS_RAW_TOOL.name
    description: str = GET_EXTENDED_LAUNCH_DATA_AS_RAW_TOOL.description
    args_schema: Type[BaseModel] = GetExtendedLaunchDataAsRawInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, launch_id: str, format: str = 'html'):
        """
        Get extended launch data as raw.

        Args:
            launch_id: Launch ID of the launch to export
            format: Format of the exported data ('pdf' or 'html')

        Returns:
            bytes: Raw content of the launch report

        Raises:
            ToolException: If export fails
        """
        try:
            response = self.client.export_specified_launch(launch_id, format)
            if not response.headers.get('Content-Disposition'):
                logger.warning(f"Exported data for launch {launch_id} is empty.")
                return None
            return response.content
        except Exception as e:
            logger.error(f"Error getting extended launch data as raw: {str(e)}")
            raise ToolException(f"Error getting extended launch data as raw: {str(e)}")


class GetLaunchDetailsTool(BaseReportPortalTool):
    """Tool to get launch details from Report Portal."""

    name: str = GET_LAUNCH_DETAILS_TOOL.name
    description: str = GET_LAUNCH_DETAILS_TOOL.description
    args_schema: Type[BaseModel] = GetLaunchDetailsInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, launch_id: str):
        """
        Get launch details.

        Args:
            launch_id: Launch ID of the launch to get details for

        Returns:
            dict: Launch details

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.get_launch_details(launch_id)
        except Exception as e:
            logger.error(f"Error getting launch details: {str(e)}")
            raise ToolException(f"Error getting launch details: {str(e)}")


class GetAllLaunchesTool(BaseReportPortalTool):
    """Tool to get all launches from Report Portal."""

    name: str = GET_ALL_LAUNCHES_TOOL.name
    description: str = GET_ALL_LAUNCHES_TOOL.description
    args_schema: Type[BaseModel] = GetAllLaunchesInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(
        self,
        page_number: int = 1,
        page_sort: str = "startTime,number,DESC",
        filter_has_composite_attribute: Optional[str] = None
    ):
        """
        Get all launches.

        Args:
            page_number: Number of page to retrieve
            page_sort: Sort order for launches
            filter_has_composite_attribute: Composite attribute filter

        Returns:
            dict: List of launches

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.get_all_launches(page_number, page_sort, filter_has_composite_attribute)
        except Exception as e:
            logger.error(f"Error getting all launches: {str(e)}")
            raise ToolException(f"Error getting all launches: {str(e)}")


class FindTestItemByIdTool(BaseReportPortalTool):
    """Tool to find test item by ID from Report Portal."""

    name: str = FIND_TEST_ITEM_BY_ID_TOOL.name
    description: str = FIND_TEST_ITEM_BY_ID_TOOL.description
    args_schema: Type[BaseModel] = FindTestItemByIdInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, item_id: str):
        """
        Find test item by ID.

        Args:
            item_id: Item ID of the item to get details for

        Returns:
            dict: Test item details

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.find_test_item_by_id(item_id)
        except Exception as e:
            logger.error(f"Error finding test item by ID: {str(e)}")
            raise ToolException(f"Error finding test item by ID: {str(e)}")


class GetTestItemsForLaunchTool(BaseReportPortalTool):
    """Tool to get test items for launch from Report Portal."""

    name: str = GET_TEST_ITEMS_FOR_LAUNCH_TOOL.name
    description: str = GET_TEST_ITEMS_FOR_LAUNCH_TOOL.description
    args_schema: Type[BaseModel] = GetTestItemsForLaunchInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, launch_id: str, page_number: int = 1, status: str = None) -> dict:
        """
        Get test items for launch.

        Args:
            launch_id: Launch ID of the launch to get test items for
            page_number: Number of page to retrieve
            status: Status of the test item

        Returns:
            dict: List of test items

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.get_test_items_for_launch(launch_id, page_number, status)
        except Exception as e:
            logger.error(f"Error getting test items for launch: {str(e)}")
            raise ToolException(f"Error getting test items for launch: {str(e)}")


class GetLogsForTestItemTool(BaseReportPortalTool):
    """Tool to get logs for test item from Report Portal."""

    name: str = GET_LOGS_FOR_TEST_ITEM_TOOL.name
    description: str = GET_LOGS_FOR_TEST_ITEM_TOOL.description
    args_schema: Type[BaseModel] = GetLogsForTestItemInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, item_id: str, page_number: int = 1):
        """
        Get logs for test item.

        Args:
            item_id: Item ID of the item to get logs for
            page_number: Number of page to retrieve

        Returns:
            dict: Test item logs

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.get_logs_for_test_items(item_id, page_number)
        except Exception as e:
            logger.error(f"Error getting logs for test item: {str(e)}")
            raise ToolException(f"Error getting logs for test item: {str(e)}")


class GetUserInformationTool(BaseReportPortalTool):
    """Tool to get user information from Report Portal."""

    name: str = GET_USER_INFORMATION_TOOL.name
    description: str = GET_USER_INFORMATION_TOOL.description
    args_schema: Type[BaseModel] = GetUserInformationInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, username: str):
        """
        Get user information.

        Args:
            username: Username of the user to get information for

        Returns:
            dict: User information

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.get_user_information(username)
        except Exception as e:
            logger.error(f"Error getting user information: {str(e)}")
            raise ToolException(f"Error getting user information: {str(e)}")


class GetDashboardDataTool(BaseReportPortalTool):
    """Tool to get dashboard data from Report Portal."""

    name: str = GET_DASHBOARD_DATA_TOOL.name
    description: str = GET_DASHBOARD_DATA_TOOL.description
    args_schema: Type[BaseModel] = GetDashboardDataInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, dashboard_id: str):
        """
        Get dashboard data.

        Args:
            dashboard_id: Dashboard ID of the dashboard to get data for

        Returns:
            dict: Dashboard data

        Raises:
            ToolException: If retrieval fails
        """
        try:
            return self.client.get_dashboard_data(dashboard_id)
        except Exception as e:
            logger.error(f"Error getting dashboard data: {str(e)}")
            raise ToolException(f"Error getting dashboard data: {str(e)}")


class UpdateTestItemTool(BaseReportPortalTool):
    """Tool to update test item status in Report Portal."""

    name: str = UPDATE_TEST_ITEM_TOOL.name
    description: str = UPDATE_TEST_ITEM_TOOL.description
    args_schema: Type[BaseModel] = UpdateTestItemInput

    def __init__(self, config: ReportPortalConfig):
        super().__init__(config)

    def execute(self, item_id: str, status: str, description: str = "Status updated manually"):
        """
        Update test item status.

        Args:
            item_id: ID of the test item to update
            status: New status for the test item (FAILED, PASSED, SKIPPED)
            description: Description for the status update

        Returns:
            dict: Update result

        Raises:
            ToolException: If update fails
        """
        # Validate status values
        valid_statuses = ["FAILED", "PASSED", "SKIPPED"]
        if status.upper() not in valid_statuses:
            raise ToolException(f"Invalid status '{status}'. Must be one of: {', '.join(valid_statuses)}")

        try:
            return self.client.update_test_item(item_id, status.upper(), description)
        except Exception as e:
            logger.error(f"Error updating test item: {str(e)}")
            raise ToolException(f"Error updating test item: {str(e)}")



import logging
from abc import abstractmethod
from operator import itemgetter
from typing import Type, Optional

from atlassian.bitbucket.cloud.repositories import Repository
from langchain_core.language_models import BaseChatModel
from langchain_core.output_parsers import StrOutputParser
from langchain_core.tools import ToolException
from pydantic import BaseModel, Field
from requests import HTTPError

from codemie_tools.base.codemie_tool import CodeMieTool
from codemie_tools.code.coder.diff_update_coder import update_content_by_task
from codemie_tools.git.bitbucket.tools_vars import (
    LIST_BRANCHES_TOOL,
    CREATE_GIT_BRANCH_TOOL,
    CREATE_PULL_REQUEST_TOOL,
    DELETE_FILE_TOOL,
    CREATE_FILE_TOOL,
    UPDATE_FILE_TOOL,
    UPDATE_FILE_DIFF_TOOL,
    SET_ACTIVE_BRANCH_TOOL,
    CREATE_PR_CHANGE_COMMENT,
    GET_PR_CHANGES
)
from codemie_tools.git.prompts import UPDATE_CONTENT_PROMPT
from codemie_tools.git.tools import UpdateFileGitTool
from codemie_tools.git.tools_models import ListBranchesToolInput
from codemie_tools.git.utils import GitCredentials, validate_bitbucket

logger = logging.getLogger(__name__)


class BranchInput(BaseModel):
    """Schema for operations that require a branch name as input."""
    branch_name: str = Field(description="The name of the branch, e.g. `my_branch`.")


class CreateBitbucketBranchTool(CodeMieTool):
    """Tool for creating branch in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = CREATE_GIT_BRANCH_TOOL.name
    description: str = """This tool is a wrapper for the Bitbucket API to create a new branch in the repository."""
    args_schema: Type[BaseModel] = BranchInput

    def execute(self, branch_name: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        created_branch = self.repo_wrapper.branches.create(
            name=branch_name,
            commit=self.repo_wrapper.base_branch
        )
        self.repo_wrapper.active_branch = created_branch.name
        return f"Branch '{created_branch.name}' created successfully, and set as current active branch."


class SetActiveBranchTool(CodeMieTool):
    """Tool for setting active branch in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = SET_ACTIVE_BRANCH_TOOL.name
    description: str = """
    This tool is a wrapper for setting the active branch in the Bitbucket repository, 
    similar to `git checkout <branch_name>` and `git switch -c <branch_name>`."""
    args_schema: Type[BaseModel] = BranchInput

    def execute(self, branch_name: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        try:
            self.repo_wrapper.active_branch = self.repo_wrapper.branches.get(
                branch_name).name
            return f"Switched to branch '{branch_name}'."
        except HTTPError as e:
            if e.response.status_code == 404:
                return f"Error {branch_name} does not exist,"
            else:
                return str(e)


class ListBranchesTool(CodeMieTool):
    """Tool for listing branches in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = LIST_BRANCHES_TOOL.name
    description: str = """
    This tool is a wrapper for the Bitbucket API to fetch a list of all branches in the repository. 
    It will return the name of each branch. No input parameters are required."""
    args_schema: Optional[Type[BaseModel]] = ListBranchesToolInput

    def execute(self, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        branches = [
            branch.name for branch in self.repo_wrapper.branches.each()
        ]
        if branches:
            branches_str = "\n".join(branches)
            return (
                f"Found {len(branches)} branches in the repository:"
                f"\n{branches_str}"
            )
        else:
            return "No branches found in the repository"


class CreatePRInput(BaseModel):
    pr_title: str = Field(description="Title of pull request. Maybe generated from made changes in the branch.")
    pr_body: str = Field(description="Body or description of the pull request of made changes.")
    source_branch: str = Field(description="Source branch of the pull request. Default is the default branch of the repository.")


class CreatePRTool(CodeMieTool):
    """Tool for creating pull request in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = CREATE_PULL_REQUEST_TOOL.name
    description: str = """
    This tool is a wrapper for the Git API to create a new pull request in a Bitbucket repository.
    Strictly follow and provide input parameters based on context.
    """
    args_schema: Type[BaseModel] = CreatePRInput

    def execute(self, pr_title: str, pr_body: str, source_branch: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        created_pr = self.repo_wrapper.pullrequests.create(
            pr_title,
            source_branch,
            self.repo_wrapper.base_branch,
            pr_body
        )
        return f"Successfully created PR number {created_pr.id}"


class DeleteFileInput(BaseModel):
    file_path: str = Field(
        description="""File path of file to be deleted. e.g. `src/agents/developer/tools/git/bitbucket_tools.py`.
        **IMPORTANT**: the path must not start with a slash"""
    )
    commit_message: str = Field(description="""This commit message can be customized by user 
        to provide more context about the deletion of the file.
        If user doesn't provide any message, default message will be used: 'Delete file {file_path}'""")


class DeleteFileTool(CodeMieTool):
    """Tool for deleting a file in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = DELETE_FILE_TOOL.name
    description: str = """This tool is a wrapper for the BitBucket API, useful when you need to delete a file in a BitBucket repository. 
    Simply pass in the full file path of the file you would like to delete. **IMPORTANT**: the path must not start with a slash"""
    args_schema: Type[BaseModel] = DeleteFileInput

    def execute(self, file_path: str, commit_message: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)

        self.repo_wrapper.sources.create(
            self.repo_wrapper.active_branch,
            commit_message,
            file_path
        )
        return f"Successfully deleted file {file_path}"


class CreateFileInput(BaseModel):
    file_path: str = Field(
        description="""File path of file to be created. e.g. `src/agents/developer/tools/git/bitbucket_tools.py`.
        **IMPORTANT**: the path must not start with a slash"""
    )
    file_contents: str = Field(
        description="""
        Full file content to be created. It must be without any escapes, just raw content to CREATE in GIT.
        Generate full file content for this field without any additional texts, escapes, just raw code content. 
        You MUST NOT ignore, skip or comment any details, PROVIDE FULL CONTENT including all content based on all best practices.
        """
    )
    commit_message: str = Field(
        description="""This commit message can be customized by user 
        to provide more context about the changes made in the file.
        If user doesn't provide any message, default message will be used: 'Create file {file_path}'"""
    )


class CreateFileTool(CodeMieTool):
    """Tool for creating a file in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = CREATE_FILE_TOOL.name
    description: str = """This tool is a wrapper for the Bitbucket API, useful when you need to create a file in a Bitbucket repository. 
    """
    args_schema: Type[BaseModel] = CreateFileInput

    def execute(self, file_path: str, file_contents: str, commit_message: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        self.repo_wrapper.sources.create(
            self.repo_wrapper.active_branch,
            commit_message,
            file_path_to_create_or_update=file_path,
            file_content_to_create_or_update=file_contents
        )
        return f"Successfully created file {file_path}"


class UpdateFileInput(BaseModel):
    file_path: str = Field(
        description="""File path of file to be updated. e.g. `src/agents/developer/tools/git/bitbucket_tools.py`.
        **IMPORTANT**: the path must not start with a slash"""
    )
    task_details: str = Field(description="""String. Specify detailed task description for file which must be updated 
        or provide detailed generated reference what should be done""")
    commit_message: str = Field(description="""This commit message can be customized by user 
        to provide more context about the changes made in the file.
        If user doesn't provide any message, default message will be used: 'Update file {file_path}'""")


class UpdateFileBitbucketTool(UpdateFileGitTool):
    """Tool for updating a file in Bitbucket."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    description: str = """This tool is a wrapper for the Bitbucket API, useful when you need to update a 
    file in a Bitbucket repository."""
    args_schema: Type[BaseModel] = UpdateFileInput

    def read_file(self, file_path):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        return self.repo_wrapper.sources.read_file_or_directory_contents(
            self.repo_wrapper.active_branch,
            file_path
        )

    def update_file(self, file_path, new_content, commit_message):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        self.repo_wrapper.sources.create(
            self.repo_wrapper.active_branch,
            commit_message,
            file_path_to_create_or_update=file_path,
            file_content_to_create_or_update=new_content
        )

    @abstractmethod
    def update_content(self, legacy_content, task_details):
        pass


class OpenAIUpdateFileWholeTool(UpdateFileBitbucketTool):
    name: str = UPDATE_FILE_TOOL.name
    llm_model: BaseChatModel = Field(exclude=True)

    def update_content(self, legacy_content, task_details):
        updated_content = self._chain.invoke(
            input={
                "question": task_details,
                "context": legacy_content,
            }
        )
        return self._parse_response(updated_content), ""

    def _parse_response(self, response: str):
        try:
            if response.startswith("```"):
                return response.split("\n", 1)[1]
            else:
                return response
        except Exception as e:
            logger.error(f"Error during parsing new content: {e}")
        return response

    @property
    def _chain(self):
        chain = (
            {
                "question": itemgetter("question"),
                "context": itemgetter("context"),
            }
            | UPDATE_CONTENT_PROMPT
            | self.llm_model
            | StrOutputParser()
        )
        return chain


class OpenAIUpdateFileDiffTool(UpdateFileBitbucketTool):
    name: str = UPDATE_FILE_DIFF_TOOL.name
    llm_model: BaseChatModel = Field(exclude=True)

    def update_content(self, legacy_content, task_details):
        return update_content_by_task(legacy_content, task_details, self.llm_model)


class GetPullRequestChangesInput(BaseModel):
    pr_number: str = Field(description="""GitLab Merge Request (Pull Request) number""")


class GetPullRequestChanges(CodeMieTool):
    """Tool for getting pull request changes."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = GET_PR_CHANGES.name
    description: str = """This tool is a wrapper for the Bitbucket API, useful when you need to get 
    all the changes from pull request in git diff format with added line numbers.
    """
    args_schema: Type[BaseModel] = GetPullRequestChangesInput
    handle_tool_error: bool = True

    def execute(self, pr_number: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        try:
            mr = self.repo_wrapper.pullrequests.get(pr_number)

            return f"""title: {mr.title}\ndescription: {mr.description}\n\n{mr.diff()}\n"""
        except HTTPError as e:
            if e.response.status_code == 404:
                raise ToolException(f"Merge request number {pr_number} wasn't found: {e}")


class CreatePullRequestChangeCommentInput(BaseModel):
    pr_number: str = Field(description="""GitLab Merge Request (Pull Request) number""")
    file_path: str = Field(description="""File path of the changed file""")
    line_number: int = Field(description="""Line number from the diff for a changed file""")
    comment: str = Field(description="""Comment content""")


class CreatePullRequestChangeComment(CodeMieTool):
    """Tool for creating a pull request comment."""
    repo_wrapper: Optional[Repository] = Field(exclude=True)
    credentials: GitCredentials
    name: str = CREATE_PR_CHANGE_COMMENT.name
    description: str = """This tool is a wrapper for the Bitbucket API, useful when you need to create 
    a comment on a pull request change.
    """
    args_schema: Type[BaseModel] = CreatePullRequestChangeCommentInput
    handle_tool_error: bool = True

    def execute(self, pr_number: str, file_path: str, line_number: int, comment: str, *args):
        validate_bitbucket(self.repo_wrapper, self.credentials)
        try:
            mr = self.repo_wrapper.pullrequests.get(pr_number)

            data = {
                "file_path": file_path,
                "line_number": line_number,
                "content": comment
            }

            return mr.comment(data)
        except HTTPError as e:
            if e.response.status_code == 404:
                raise ToolException(f"Merge request number {pr_number} wasn't found: {e}")
            else:
                raise e



import logging
import re
from abc import abstractmethod
from operator import itemgetter
from typing import Type, Dict, Union, Optional

from gitlab.exceptions import GitlabGetError
from langchain_core.language_models import BaseChatModel
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.tools import ToolException
from pydantic import BaseModel, Field

from codemie_tools.base.codemie_tool import CodeMieTool
from codemie_tools.base.utils import parse_tool_input
from codemie_tools.code.coder.diff_update_coder import update_content_by_task
from codemie_tools.git.gitlab.custom_gitlab_api_wrapper import CustomGitLabAPIWrapper
from codemie_tools.git.gitlab.tools_vars import UPDATE_FILE_TOOL, UPDATE_FILE_DIFF_TOOL
from codemie_tools.git.prompts import UPDATE_CONTENT_PROMPT
from codemie_tools.git.tools import UpdateFileGitTool
from codemie_tools.git.tools_models import ListBranchesToolInput
from codemie_tools.git.utils import validate_gitlab_wrapper, GitCredentials

logger = logging.getLogger(__name__)


class BranchInput(BaseModel):
    """Schema for operations that require a branch name as input."""

    branch_name: str = Field(description="The name of the branch, e.g. `my_branch`.")


class CreateGitLabBranchTool(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "create_branch"
    description: str = """This tool is a wrapper for the GitLab API to create a new branch in the repository."""
    args_schema: Type[BaseModel] = BranchInput

    def execute(self, branch_name: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.create_branch(branch_name)


class CreatePRInput(BaseModel):
    pr_title: str = Field(description="Title of pull request. Maybe generated from made changes in the branch.")
    pr_body: str = Field(description="Body or description of the pull request of made changes.")


class CreatePRTool(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "create_pull_request"
    description: str = """This tool is a wrapper for the GitLab API to create a new pull request in a GitLab repository.
    Strictly follow and provide input parameters based on context.
    """
    args_schema: Type[BaseModel] = CreatePRInput

    def execute(self, pr_title: str, pr_body: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.create_pull_request(pr_title + "\n\n" + pr_body)


class DeleteFileInput(BaseModel):
    file_path: str = Field(description="""File path of file to be deleted. e.g. `src/agents/developer/tools/git/github_tools.py`.
    **IMPORTANT**: the path must not start with a slash""")
    commit_message: str = Field(description="""This commit message can be customized by user 
    to provide more context about the deletion of the file.
    If user doesn't provide any message, default message will be used: 'Delete file {file_path}'""")


class DeleteFileTool(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "delete_file"
    description: str = """This tool is a wrapper for the GitLab API, useful when you need to delete a file in a GitLab repository. 
    Simply pass in the full file path of the file you would like to delete. **IMPORTANT**: the path must not start with a slash"""
    args_schema: Type[BaseModel] = DeleteFileInput

    def execute(self, file_path: str, commit_message: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.delete_file(file_path, commit_message)


class CreateFileInput(BaseModel):
    file_path: str = Field(description="""File path of file to be created. e.g. `src/agents/developer/tools/git/github_tools.py`.
    **IMPORTANT**: the path must not start with a slash""")
    file_contents: str = Field(description="""
    Full file content to be created. It must be without any escapes, just raw content to CREATE in GIT.
    Generate full file content for this field without any additional texts, escapes, just raw code content. 
    You MUST NOT ignore, skip or comment any details, PROVIDE FULL CONTENT including all content based on all best practices.
    """)
    commit_message: str = Field(description="""This commit message can be customized by user 
    to provide more context about the changes made in the file.
    If user doesn't provide any message, default message will be used: 'Create file {file_path}'""")


class CreateFileTool(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "create_file"
    description: str = """This tool is a wrapper for the GitLab API, useful when you need to create a file in a GitLab repository.
    """
    args_schema: Type[BaseModel] = CreateFileInput

    def execute(self, file_path: str, file_contents: str, commit_message: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.create_file(file_path + "\n " + file_contents, commit_message)


class SetActiveBranchTool(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "set_active_branch"
    description: str = """
    This tool is a wrapper for the Git API and set the active branch in the repository, similar to `git checkout <branch_name>` and `git switch -c <branch_name>`."""
    args_schema: Type[BaseModel] = BranchInput

    def execute(self, branch_name: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.set_active_branch(branch_name=branch_name)


class ListBranchesTool(CodeMieTool):
    """Tool for interacting with the GitHub API."""

    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "list_branches_in_repo"
    description: str = """This tool is a wrapper for the Git API to fetch a list of all branches in the repository. 
    It will return the name of each branch. No input parameters are required."""
    args_schema: Optional[Type[BaseModel]] = ListBranchesToolInput

    def execute(self, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.list_branches_in_repo()


class UpdateFileInput(BaseModel):
    file_path: str = Field(description="""File path of file to be updated. e.g. `src/agents/developer/tools/git/github_tools.py`.
    **IMPORTANT**: the path must not start with a slash""")
    task_details: str = Field(description="""String. Specify detailed task description for 
    file which must be updated or provide detailed generated reference what should be done""")
    commit_message: str = Field(description="""This commit message can be customized by user 
    to provide more context about the changes made in the file.
    If user doesn't provide any message, default message will be used: 'Update file {file_path}'""")


class UpdateFileGitLabTool(UpdateFileGitTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    description: str = """This tool is a wrapper for the GitHub API, useful when you need to update a file in a 
    GitHub repository."""
    args_schema: Type[BaseModel] = UpdateFileInput

    def read_file(self, file_path):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        return self.api_wrapper.read_file(file_path)

    def update_file(self, file_path, new_content, commit_message):
        self.api_wrapper.replace_file_content(file_path + "\n" + new_content, commit_message)

    @abstractmethod
    def update_content(self, legacy_content, task_details):
        pass


class OpenAIUpdateFileWholeTool(UpdateFileGitLabTool):
    """Tool for interacting with the GitHub API."""

    name: str = UPDATE_FILE_TOOL.name
    description: str = UPDATE_FILE_TOOL.description
    llm_model: BaseChatModel = Field(exclude=True)
    update_prompt: PromptTemplate = UPDATE_CONTENT_PROMPT

    def update_content(self, legacy_content, task_details):
        updated_content = self._chain.invoke(
            input={
                "question": task_details,
                "context": legacy_content,
            }
        )

        return self._parse_response(updated_content), ""

    def _parse_input(self, tool_input: Union[str, Dict], tool_call_id: Optional[str], *args, **kwargs):
        return parse_tool_input(self.args_schema, tool_input)

    def _parse_response(self, response: str):
        try:
            if response.startswith("```"):
                return response.split("\n", 1)[1]
            else:
                return response
        except Exception as e:
            logger.error(f"Error during parsing new content: {e}")
        return response

    @property
    def _chain(self):
        chain = (
                {
                    "question": itemgetter("question"),
                    "context": itemgetter("context"),
                }
                | self.update_prompt
                | self.llm_model
                | StrOutputParser()
        )
        return chain


class OpenAIUpdateFileDiffTool(UpdateFileGitLabTool):
    """Tool for interacting with the GitHub API."""

    name: str = UPDATE_FILE_DIFF_TOOL.name
    description: str = UPDATE_FILE_DIFF_TOOL.description
    llm_model: BaseChatModel = Field(exclude=True)

    def update_content(self, legacy_content, task_details):
        return update_content_by_task(legacy_content, task_details, self.llm_model)


def get_position(line_number, file_path, mr):
    changes = mr.changes()["changes"]
    # Get first change 
    change = next((item for item in changes if item.get("new_path") == file_path), None)
    if change is None:
        change = next((item for item in changes if item.get("old_path") == file_path), None)
    if change is None:
        raise ValueError(f"Change for file {file_path} wasn't found in PR")

    position = get_diff_w_position(change=change)[line_number][0]

    position.update({
        "base_sha": mr.diff_refs["base_sha"],
        "head_sha": mr.diff_refs["head_sha"],
        "start_sha": mr.diff_refs["start_sha"],
        'position_type': 'text'
    })

    return position


def get_diff_w_position(change):
    diff = change["diff"]
    diff_with_ln = {}
    # Regular expression to extract old and new line numbers
    pattern = r"^@@ -(\d+),(\d+) \+(\d+),(\d+) @@"
    # GitLab API requires new path and line for added lines, old path and files for removed lines.
    # For unchaged lines it requires both. 
    for index, line in enumerate(diff.split("\n")):
        position = {}
        match = re.match(pattern, line)
        if match:
            old_line = int(match.group(1))
            new_line = int(match.group(3))
        elif line.startswith("+"):
            position["new_line"] = new_line
            position["new_path"] = change["new_path"]
            new_line += 1
        elif line.startswith("-"):
            position["old_line"] = old_line
            position["old_path"] = change["old_path"]
            old_line += 1
        elif line.startswith(" "):
            position["old_line"] = old_line
            position["old_path"] = change["old_path"]
            position["new_line"] = new_line
            position["new_path"] = change["new_path"]
            new_line += 1
            old_line += 1
        elif line.startswith("\\"):
            # Assign previos position to \\ metadata
            position = diff_with_ln[index - 1][0]
        else:
            # Stop at final empty line
            break

        diff_with_ln[index] = [position, line]

        # Assign next position to @@ metadata
        if index > 0 and diff_with_ln[index - 1][1].startswith("@"):
            diff_with_ln[index - 1][0] = position

    return diff_with_ln


class GetPullRequesChangesInput(BaseModel):
    pr_number: str = Field(description="""GitLab Merge Request (Pull Request) number""")


class GetPullRequesChanges(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "get_pr_changes"
    description: str = """This tool is a wrapper for the GitLab API, useful when you need to get all the changes from pull request in git diff format with added line numbers.
    """
    args_schema: Type[BaseModel] = GetPullRequesChangesInput
    handle_tool_error: bool = True

    def execute(self, pr_number: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        try:
            repo = self.api_wrapper.gitlab_repo_instance
            try:
                mr = repo.mergerequests.get(pr_number)
            except GitlabGetError as e:
                if e.response_code == 404:
                    raise ToolException(f"Merge request number {pr_number} wasn't found: {e}")

            res = f"""title: {mr.title}\ndescription: {mr.description}\n\n"""

            for change in mr.changes()["changes"]:
                diff_w_position = get_diff_w_position(change=change)
                diff = "\n".join([str(line_num) + ":" + line[1] for line_num, line in diff_w_position.items()])

                res = res + f"""diff --git a/{change["old_path"]} b/{change["new_path"]}\n{diff}\n"""

            return res
        except ToolException:
            raise
        except Exception as e:
            raise ToolException(f"An error occurred: {e}")


class CreatePullRequestChangeCommentInput(BaseModel):
    pr_number: str = Field(description="""GitLab Merge Request (Pull Request) number""")
    file_path: str = Field(description="""File path of the changed file""")
    line_number: int = Field(description="""Line number from the diff for a changed file""")
    comment: str = Field(description="""Comment content""")


class CreatePullRequestChangeComment(CodeMieTool):
    api_wrapper: Optional[CustomGitLabAPIWrapper] = Field(exclude=True, default=None)
    credentials: GitCredentials
    name: str = "create_pr_change_comment"
    description: str = """This tool is a wrapper for the GitLab API, useful when you need to create a comment on a pull request change.
    """
    args_schema: Type[BaseModel] = CreatePullRequestChangeCommentInput
    handle_tool_error: bool = True

    def execute(self, pr_number: str, file_path: str, line_number: int, comment: str, *args):
        validate_gitlab_wrapper(self.api_wrapper, self.credentials)
        repo = self.api_wrapper.gitlab_repo_instance
        try:
            mr = repo.mergerequests.get(pr_number)
        except GitlabGetError as e:
            if e.response_code == 404:
                raise ToolException(f"Merge request number {pr_number} wasn't found: {e}")
        try:
            position = get_position(file_path=file_path, line_number=line_number, mr=mr)

            mr.discussions.create({"body": comment, "position": position})
            return "Comment added"
        except Exception as e:
            raise ToolException(f"An error occurred: {e}")



from typing import List

from codemie_tools.base.models import ToolKit, ToolSet, Tool
from .azure_devops_git.tools import AzureDevOpsGitTool
from .azure_devops_git.tools_vars import AZURE_DEVOPS_GIT_TOOL
from .github.tools import GithubTool
from .github.tools_vars import GITHUB_TOOL
from .gitlab.tools import GitlabTool
from .gitlab.tools_vars import GITLAB_TOOL
from ...base.base_toolkit import DiscoverableToolkit


class VcsToolkitUI(ToolKit):
    toolkit: ToolSet = ToolSet.VCS
    tools: List[Tool] = [
        Tool.from_metadata(AZURE_DEVOPS_GIT_TOOL, tool_class=AzureDevOpsGitTool),
        Tool.from_metadata(GITHUB_TOOL, tool_class=GithubTool),
        Tool.from_metadata(GITLAB_TOOL, tool_class=GitlabTool),
    ]


class VcsToolkit(DiscoverableToolkit):

    @classmethod
    def get_definition(cls):
        return VcsToolkitUI()



from codemie_tools.base.models import ToolMetadata
from codemie_tools.qa.zephyr.models import ZephyrConfig

ZEPHYR_TOOL = ToolMetadata(
    name="ZephyrScale",
    description=
    """
    Zephyr Tool that provides access to the zephyr-python-api package, enabling interaction with Zephyr test cases,
    cycles or executions etc.
    You must provide the following args: entity_str, method_str and body.
    1. 'entity_str': ZephyrScale entity requested by user. It must be accessible from ZephyrScale().api
    2. 'method_str': Zephyr method that is accessible from that entity. E.g. method 'get_test_cases' is accessible from ZephyrScale().api.test_cases.
        It can be equal to 'dir' so that you can list all available methods in requested entity.
    3. 'body': (Optional) Valid JSON object with parameters that must be passed to method.
    If some required information is not provided by user, try find by querying API, if not found ask user.
    If method you requested is not exists, try to execute tool with method 'dir' to get list of available methods.
    """.strip(),
    label="Zephyr Scale",
    user_description="""
    Provides access to the zephyr-python-api package, enabling interaction with Zephyr test cases, cycles or executions.
    Before using it, it is necessary to add a new integration for the tool by providing:
    1. Alias (A friendly name for the Zephyr Scale integration)
    2. Url (base url of the Zephyr Scale)
    3. Token (API access token)
    Usage Note:
    Use this tool when you need to manage Zephyr test cases, cycles or executions.
    """.strip(),
    settings_config=True,
    config_class=ZephyrConfig
)



import logging
from typing import BinaryIO, Any

import markitdown
from markitdown import StreamInfo, DocumentConverterResult

from codemie_tools.file_analysis.xlsx.processor import XlsxProcessor

logger = logging.getLogger(__name__)


class XlsxConverter(markitdown.converters.XlsxConverter):
    """
    Converts XLSX files to Markdown, with each sheet presented as a separate Markdown table.
    Customized to filter out empty rows and columns (NaN values) to reduce markdown size and improve readability.
    Can filter sheets based on visibility.
    """

    def __init__(self, sheet_names: list[str]=None, visible_only: bool=True) -> None:
        super().__init__()
        self.processor = XlsxProcessor(sheet_names=sheet_names, visible_only=visible_only)
        logger.info("Enable custom XLSX converter")

    def convert(
            self,
            file_stream: BinaryIO,
            stream_info: StreamInfo,
            **kwargs: Any,  # Options to pass to the converter
    ) -> DocumentConverterResult:
        """
        Converts Excel files to markdown tables, filtering out empty rows and columns to reduce output size.
        If visible_only is True, only visible sheets will be processed.
        """

        # Use the processor to load and clean Excel data
        sheets_clean = self.processor.load(file_stream, clean_data=True)

        # Use the processor to convert the sheets to markdown
        md_content = self.processor.convert(sheets_clean, **kwargs)

        return DocumentConverterResult(markdown=md_content)


from typing import Any

import boto3
from botocore.config import Config
from langchain_core.tools import ToolException


class AWSClient:
    """HTTP client for AWS API using boto3."""

    def __init__(self, region: str, access_key_id: str, secret_access_key: str, session_token: str | None = None):
        """
        Initialize the AWS client.

        Args:
            region: AWS region name
            access_key_id: AWS access key ID
            secret_access_key: AWS secret access key
        """
        self.region = region
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.session_token = session_token

    def get_client(self, service: str):
        """
        Get a boto3 client for the specified AWS service.

        Args:
            service: AWS service name (e.g., 'ec2', 'iam', 's3')

        Returns:
            boto3 client instance for the service

        Raises:
            ToolException: If client creation fails
        """
        try:
            client_config = Config(region_name=self.region)
            client = boto3.client(
                service,
                aws_access_key_id=self.access_key_id,
                aws_secret_access_key=self.secret_access_key,
                aws_session_token=self.session_token,
                config=client_config,
            )
            return client
        except Exception as e:
            raise ToolException(f"Failed to create AWS client for service '{service}': {str(e)}")

    def execute_method(self, service: str, method_name: str, method_arguments: dict) -> Any:
        """
        Execute a method on an AWS service client.

        Args:
            service: AWS service name
            method_name: The method to call on the client
            method_arguments: Arguments to pass to the method

        Returns:
            The response from the AWS API

        Raises:
            ToolException: If method execution fails
        """
        try:
            client = self.get_client(service)

            if not hasattr(client, method_name):
                raise ToolException(
                    f"Method '{method_name}' does not exist for AWS service '{service}'"
                )

            method = getattr(client, method_name)
            response = method(**method_arguments)
            return response

        except ToolException:
            raise
        except Exception as e:
            raise ToolException(
                f"Failed to execute {service}.{method_name}: {str(e)}"
            )

    def health_check(self):
        """
        Check if AWS credentials are valid.
        Uses STS, which works universally for all credential types.
        """
        client = self.get_client("sts")
        client.get_caller_identity()



from codemie_tools.base.models import ToolMetadata
from .models import ServiceNowConfig

SNOW_TABLE_TOOL = ToolMetadata(
    name="servicenow_table_tool",
    description="""
    ServiceNow Tool for Official ServiceNow Table REST API call, searching, creating, updating table, etc. 
    You must provide the following args: relative_url, method, params. 
    1. 'method': The HTTP method, e.g. 'GET', 'POST', 'PUT', 'DELETE' etc. Some of them might be turned off in configuration.
    2. 'table': The name of the table to work with
    3. 'sys_id': Optional, used when working with a specific record, rather than entire table. In this case the api in use
    will be /api/now/table/{tableName}/{sys_id}
    4. 'query': Optional set of query parameters to be used if supported by API. f.ex: `sysparm_query`, `sysparm_offset`, `sysparm_limit`, etc. In a form of JSON.
    """,
    label="ServiceNow Table API",
    user_description="""
    Provides access to the ServiceNow Table API.
    """.strip(),
    settings_config=True,
    config_class=ServiceNowConfig
)


from typing import Optional, Dict, Any, Union, List

from pydantic import BaseModel, Field, model_validator

from codemie_tools.base.models import CodeMieToolConfig, CredentialTypes, RequiredField


class GCPConfig(CodeMieToolConfig):
    """Configuration for GCP integration."""
    credential_type: CredentialTypes = Field(default=CredentialTypes.GCP, exclude=True, frozen=True)

    service_account_key: str = RequiredField(
        description="GCP Service Account Key in JSON format",
        json_schema_extra={
            "placeholder": '{"type": "service_account", "project_id": "...", "private_key_id": "...", ...}',
            "sensitive": True,
            "help": "https://cloud.google.com/iam/docs/creating-managing-service-account-keys"
        }
    )

    @model_validator(mode='before')
    def validate_config(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Support legacy credential keys for backward compatibility."""
        # Map legacy key to new key
        if "gcp_api_key" in values:
            values["service_account_key"] = values.pop("gcp_api_key")

        return values


class GCPInput(BaseModel):
    """Input schema for GCP tool operations."""

    method: str = Field(
        description="""
        The HTTP method to use for the request.
        Supported methods: GET, POST, PUT, DELETE, PATCH

        Example: "GET", "POST"
        """
    )

    scopes: List[str] = Field(
        description="""
        List of OAuth 2.0 scopes for Google APIs.
        Required for authentication and authorization.

        Common scopes:
        - https://www.googleapis.com/auth/cloud-platform (Full access to all Google Cloud resources)
        - https://www.googleapis.com/auth/compute (Google Compute Engine)
        - https://www.googleapis.com/auth/devstorage.read_write (Cloud Storage)
        - https://www.googleapis.com/auth/logging.admin (Cloud Logging)

        Example: ["https://www.googleapis.com/auth/cloud-platform"]
        """
    )

    url: str = Field(
        description="""
        Absolute URI for Google Cloud REST API.
        Must be a valid googleapis.com endpoint.

        Example: "https://compute.googleapis.com/compute/v1/projects/{project}/zones/{zone}/instances"
        Example: "https://storage.googleapis.com/storage/v1/b/{bucket}/o"
        """
    )

    optional_args: Optional[Union[str, Dict[str, Any]]] = Field(
        default=None,
        description="""
        Optional JSON object containing additional request parameters.
        Possible keys: 'data', 'json', 'params', 'headers'

        Example: {"params": {"maxResults": 100}, "headers": {"Content-Type": "application/json"}}

        Note: Must be valid JSON. No comments allowed.
        """
    )



import logging
import re
import time
import uuid
from typing import Tuple, List

from langchain_core.messages import AIMessage, HumanMessage

from codemie_tools.code.coder.diff_update_prompts import base_messages
from codemie_tools.code.linter.facade import LinterFacade

logger = logging.getLogger(__name__)


def update_content_by_task(old_content, task_details, llm):
    origin_temp, origin_top_p = set_llm_params(llm)
    try:
        return solve_task_with_retry(old_content, task_details, llm)
    finally:
        restore_llm_params(llm, origin_temp, origin_top_p)


def solve_task_with_retry(old_content, task_details, llm):
    messages = base_messages + [
        HumanMessage(content=old_content),
        AIMessage(content="Ok. I will focus on making changes in the current file content"),
        HumanMessage(content=f"Here it the task: {task_details}")
    ]
    return call_and_process_llm(llm, messages, old_content)


MAX_RETRIES = 3
BASE_DELAY = 1  # in seconds


def call_and_process_llm(llm, messages, old_content):
    # Generate a unique ID for this method call
    call_id = str(uuid.uuid4())[:8]  # Using first 8 characters of a UUID for brevity

    for attempt in range(MAX_RETRIES):
        llm_response = llm.invoke(messages)
        try:
            new_content, edits = extract_and_apply_edits(llm_response.content, old_content)
            return new_content, pretty_format_edits(edits)
        except ValueError as e:
            delay = BASE_DELAY * (2 ** attempt)
            logger.warning(
                f"Diff coder. Error while processing llm response for call {call_id}."
                f" Error: {str(e)}"
                f" Attempt {attempt + 1} failed. Retrying in {delay} seconds...")
            if attempt == MAX_RETRIES - 1:
                raise e
            messages = messages + [
                llm_response,
                HumanMessage(content=f"I got the following error processing your response: \"{str(e)}\". Fix it")
            ]
            time.sleep(delay)


def pretty_format_edits(edits: List[Tuple[str, str]]) -> str:
    return "\n".join(
        [f"Change {i + 1}:\nOriginal Code:\n{original}\n\nNew Code:\n{new}\n{'-' * 40}" for i, (original, new) in
         enumerate(edits)]
    )

def set_llm_params(llm):
    if (temperature := getattr(llm, 'temperature', None)) is not None:
        setattr(llm, 'temperature', 0.1)
    if (top_p := getattr(llm, 'top_p', None)) is not None:
        setattr(llm, 'top_p', 0.2)
    return temperature, top_p


def extract_and_apply_edits(llm_response, old_content):
    edits = get_edits(llm_response)
    if not edits:
        raise ValueError("There are no *SEARCH/REPLACE* blocks in the response")

    content_candidate = apply_edits(edits, old_content)
    suc, errors = LinterFacade().lint_code(
        get_lang_from_response(llm_response),
        old_content,
        content_candidate)
    if not suc:
        logger.warning(f"Code check failed Errors:{errors}")
        raise ValueError(f"Code check failed. Errors:\n{errors}")
    return content_candidate, edits


def get_lang_from_response(llm_response: str) -> str:
    match = re.search(r'!!!(\w+)', llm_response)
    if match:
        return match.group(1)
    else:
        logger.info("Language not found in the LLM response.")
        return ""


def restore_llm_params(llm, origin_temp, origin_top_p):
    if hasattr(llm, 'temperature'):
        setattr(llm, 'temperature', origin_temp)
    if hasattr(llm, 'top_p'):
        setattr(llm, 'top_p', origin_top_p)


def get_edits(content):
    return list(find_original_update_blocks(content))


def apply_edits(edits, content):
    failed = []
    passed = []
    for edit in edits:
        original, updated = edit
        new_content = do_replace(content, original, updated)
        if new_content:
            content = new_content
            passed.append(edit)
        else:
            failed.append(edit)

    if not failed:
        return content

    blocks = "block" if len(failed) == 1 else "blocks"

    res = f"# {len(failed)} SEARCH/REPLACE {blocks} failed to match!\n"
    for edit in failed:
        original, updated = edit

        res += f"""
## SearchReplaceNoExactMatch: This SEARCH block failed to exactly match lines
<<<<<<< SEARCH
{original}=======
{updated}>>>>>>> REPLACE

The SEARCH section must exactly match an existing block of lines including all white space, comments, indentation,
docstrings, etc
"""
    raise ValueError(res)


def find_original_update_blocks(content):
    # make sure we end with a newline, otherwise the regex will miss <<UPD on the last line
    if not content.endswith("\n"):
        content = content + "\n"

    head = "<<<<<<< SEARCH"
    divider = "======="
    updated = ">>>>>>> REPLACE"

    separators = "|".join([head, divider, updated])

    split_re = re.compile(r"^((?:" + separators + r")[ ]*\n)", re.MULTILINE | re.DOTALL)

    pieces = re.split(split_re, content)

    pieces.reverse()
    processed = []

    try:
        while pieces:
            cur = pieces.pop()

            if cur.strip() != head:
                processed.append(cur)
                continue

            processed.append(cur)  # original_marker

            original_text = pieces.pop()
            processed.append(original_text)

            divider_marker = pieces.pop()
            processed.append(divider_marker)
            if divider_marker.strip() != divider:
                raise ValueError(f"Expected `{divider}` not {divider_marker.strip()}")

            updated_text = pieces.pop()
            processed.append(updated_text)

            updated_marker = pieces.pop()
            processed.append(updated_marker)
            if updated_marker.strip() != updated:
                raise ValueError(f"Expected `{updated}` not `{updated_marker.strip()}")

            yield original_text, updated_text
    except ValueError as e:
        processed = "".join(processed)
        err = e.args[0]
        raise ValueError(f"{processed}\n^^^ {err}")
    except IndexError:
        processed = "".join(processed)
        raise ValueError(f"{processed}\n^^^ Incomplete SEARCH/REPLACE block.")
    except Exception:
        processed = "".join(processed)
        raise ValueError(f"{processed}\n^^^ Error parsing SEARCH/REPLACE block.")


def do_replace(content, before_text, after_text):
    before_text = strip_quoted_wrapping(before_text)

    after_text = strip_quoted_wrapping(after_text)

    if content is None:
        return

    if not before_text.strip():
        # append to existing file, or start a new file
        new_content = content + after_text
    else:
        new_content = replace_most_similar_chunk(content, before_text, after_text)

    return new_content


def strip_quoted_wrapping(res):
    if not res:
        return res

    res = res.splitlines()

    if res[0].startswith("!!!") and res[-1].startswith("!!!"):
        res = res[1:-1]

    res = "\n".join(res)
    if res and res[-1] != "\n":
        res += "\n"

    return res


def replace_most_similar_chunk(whole, part, replace):
    """Best efforts to find the `part` lines in `whole` and replace them with `replace`"""
    occur_count = whole.count(part)
    if occur_count > 1:
        raise ValueError(f"The following SEARCH block is not unique in the original content: \n\"{part}\"")

    whole, whole_lines = prep(whole)
    part, part_lines = prep(part)
    replace, replace_lines = prep(replace)

    res = perfect_or_whitespace(whole_lines, part_lines, replace_lines)
    if res:
        return res

    # drop leading empty line, GPT sometimes adds them spuriously¬
    if len(part_lines) > 2 and not part_lines[0].strip():
        skip_blank_line_part_lines = part_lines[1:]
        res = perfect_or_whitespace(whole_lines, skip_blank_line_part_lines, replace_lines)
        if res:
            return res

    # Try to handle when it elides code with ...
    try:
        res = try_dotdotdots(whole, part, replace)
        if res:
            return res
    except ValueError:
        pass


def prep(content):
    if content and not content.endswith("\n"):
        content += "\n"
    lines = content.splitlines(keepends=True)
    return content, lines


def perfect_or_whitespace(whole_lines, part_lines, replace_lines):
    # Try for a perfect match
    res = perfect_replace(whole_lines, part_lines, replace_lines)
    if res:
        return res

    # Try being flexible about leading whitespace
    res = replace_part_with_missing_leading_whitespace(whole_lines, part_lines, replace_lines)
    if res:
        return res


def perfect_replace(whole_lines, part_lines, replace_lines):
    part_tup = tuple(part_lines)
    part_len = len(part_lines)

    for i in range(len(whole_lines) - part_len + 1):
        whole_tup = tuple(whole_lines[i: i + part_len])
        if part_tup == whole_tup:
            res = whole_lines[:i] + replace_lines + whole_lines[i + part_len:]
            return "".join(res)


def replace_part_with_missing_leading_whitespace(whole_lines, part_lines, replace_lines):
    # GPT often messes up leading whitespace.
    # It usually does it uniformly across the ORIG and UPD blocks.
    # Either omitting all leading whitespace, or including only some of it.

    # Outdent everything in part_lines and replace_lines by the max fixed amount possible
    leading = [len(p) - len(p.lstrip()) for p in part_lines if p.strip()] + [
        len(p) - len(p.lstrip()) for p in replace_lines if p.strip()
    ]

    if leading and min(leading):
        num_leading = min(leading)
        part_lines = [p[num_leading:] if p.strip() else p for p in part_lines]
        replace_lines = [p[num_leading:] if p.strip() else p for p in replace_lines]

    # can we find an exact match not including the leading whitespace
    num_part_lines = len(part_lines)

    for i in range(len(whole_lines) - num_part_lines + 1):
        add_leading = match_but_for_leading_whitespace(
            whole_lines[i: i + num_part_lines], part_lines
        )

        if add_leading is None:
            continue

        replace_lines = [add_leading + rline if rline.strip() else rline for rline in replace_lines]
        whole_lines = whole_lines[:i] + replace_lines + whole_lines[i + num_part_lines:]
        return "".join(whole_lines)

    return None


def match_but_for_leading_whitespace(whole_lines, part_lines):
    num = len(whole_lines)

    # does the non-whitespace all agree?
    if not all(whole_lines[i].lstrip() == part_lines[i].lstrip() for i in range(num)):
        return

    # are they all offset the same?
    add = set(
        whole_lines[i][: len(whole_lines[i]) - len(part_lines[i])]
        for i in range(num)
        if whole_lines[i].strip()
    )

    if len(add) != 1:
        return

    return add.pop()


def try_dotdotdots(whole, part, replace):
    """
    See if the edit block has ... lines.
    If not, return none.

    If yes, try and do a perfect edit with the ... chunks.
    If there's a mismatch or otherwise imperfect edit, raise ValueError.

    If perfect edit succeeds, return the updated whole.
    """

    dots_re = re.compile(r"(^\s*\.\.\.\n)", re.MULTILINE | re.DOTALL)

    part_pieces = re.split(dots_re, part)
    replace_pieces = re.split(dots_re, replace)

    if not validate_pieces(part_pieces, replace_pieces):
        return

    part_pieces = [part_pieces[i] for i in range(0, len(part_pieces), 2)]
    replace_pieces = [replace_pieces[i] for i in range(0, len(replace_pieces), 2)]

    pairs = zip(part_pieces, replace_pieces)
    for part, replace in pairs:
        if not part and not replace:
            continue
        if not part and replace:
            if not whole.endswith("\n"):
                whole += "\n"
            whole += replace
            continue
        if whole.count(part) != 1:
            raise ValueError
        whole = whole.replace(part, replace, 1)

    return whole


def validate_pieces(part_pieces, replace_pieces):
    if len(part_pieces) != len(replace_pieces):
        raise ValueError("Unpaired ... in SEARCH/REPLACE block")
    if len(part_pieces) == 1:
        return False
    all_dots_match = all(part_pieces[i] == replace_pieces[i] for i in range(1, len(part_pieces), 2))
    if not all_dots_match:
        raise ValueError("Unmatched ... in SEARCH/REPLACE block")
    return True



import logging
import subprocess
from typing import Tuple, Dict, List

from codemie_tools.code.linter.linter import Linter

logger = logging.getLogger(__name__)


class PythonLinter(Linter):
    error_codes: str

    def __init__(self, error_codes):
        self.error_codes = error_codes

    def lint_code_diff(self, old_content: str, new_content: str) -> Tuple[bool, str]:
        errors = self.lint_single_code(new_content)

        if not errors:
            return True, ""

        changed_lines = self.get_changed_lines(old_content, new_content)
        new_errors = self._filter_new_errors(errors, changed_lines)

        if not new_errors:
            return True, ""

        error_message = self._format_errors(new_errors)
        return False, error_message

    def lint_single_code(self, content: str) -> Dict[int, str]:
        try:
            error_lines = self._run_flake8_cli(content, self.error_codes)
        except subprocess.SubprocessError as e:
            logger.error(f"An error occurred while running flake8: {e}")
            return {}

        error_dict = {}
        for error in error_lines:
            parts = error.split(':')
            if len(parts) > 1:
                try:
                    line_number = int(parts[1])
                    error_dict[line_number] = error
                except ValueError:
                    logger.error(f"Could not parse line number from error: {error}")

        return error_dict

    @staticmethod
    def _run_flake8_cli(content: str, error_codes: str):
        res_cli = subprocess.run(
            ["flake8", "--select", error_codes, "-"],
            input=content,
            text=True,
            capture_output=True,
            check=False
        )
        return res_cli.stdout.splitlines()

    @staticmethod
    def _filter_new_errors(errors: Dict[int, str], changed_lines: Dict[int, str]) -> List[Tuple[str, str]]:
        retained_errors = []

        for line_number, error_detail in errors.items():
            if line_number in changed_lines:
                line_content = changed_lines[line_number]
                retained_errors.append((line_content, error_detail))

        return retained_errors

    @staticmethod
    def _format_errors(errors: List[Tuple[str, str]]) -> str:
        formatted_errors = []
        for line_content, error_details in errors:
            formatted_errors.append(f"Line: {line_content}\nError: {error_details}\n")
        return "\n".join(formatted_errors)



import json
from typing import Optional, Dict, Any, Type

import requests
from pydantic import BaseModel, Field

from codemie_tools.base.codemie_tool import CodeMieTool
from codemie_tools.notification.telegram.models import TelegramConfig
from codemie_tools.notification.telegram.tools_vars import TELEGRAM_TOOL


class TelegramToolInput(BaseModel):
    method: str = Field(
        ..., description="The HTTP method to use for the request (GET, POST). Required parameter."
    )
    relative_url: str = Field(
        ...,
        description="""
        The relative URL of the Telegram Bot API to call, e.g. 'sendMessage'. Required parameter.
        In case of GET method, you MUST include query parameters in the URL.
        """,
    )
    params: Optional[str] = Field(
        default="",
        description="""
        Optional JSON of parameters to be sent in request body or query params. MUST be string with valid JSON. 
        Important: to send message, you MUST get "chat_id" parameter first.
        """,
    )


class TelegramTool(CodeMieTool):
    config: TelegramConfig
    name: str = TELEGRAM_TOOL.name
    description: str = TELEGRAM_TOOL.description
    args_schema: Type[BaseModel] = TelegramToolInput

    def execute(self, method: str, relative_url: str, params: Optional[str] = "") -> str:
        """
        Execute a Telegram Bot API request.

        Args:
            method: The HTTP method to use (GET, POST, etc.)
            relative_url: The relative URL of the Telegram API endpoint
            params: JSON string of parameters to send

        Returns:
            Response from the Telegram API
        """
        if not relative_url.startswith("/"):
            relative_url = f"/{relative_url}"

        # If config is None, raise a meaningful error
        if self.config is None:
            raise ValueError(
                "Telegram config is provided set. Please set it before using the tool."
            )

        if not self.config.bot_token:
            raise ValueError(
                "Telegram token is not set. Please provide it before using the tool."
            )

        base_url = f"https://api.telegram.org/bot{self.config.bot_token}"
        full_url = f"{base_url}{relative_url}"
        headers = {"Content-Type": "application/json"}
        payload_params = self._parse_payload_params(params)
        response = requests.request(method, full_url, headers=headers, json=payload_params)
        response.raise_for_status()
        return response.text

    def _parse_payload_params(self, params: Optional[str]) -> Dict[str, Any]:
        """Parse JSON string into dictionary for request parameters"""
        if params:
            json_acceptable_string = params.replace("'", '"')
            return json.loads(json_acceptable_string)
        return {}

    def _healthcheck(self):
        """
        Check if the Telegram Bot API token is valid.

        Returns:
            Tuple of (success: bool, error_message: str)
        """
        if self.config is None:
            raise ValueError("Telegram config is not set. Please set it before using the tool.")
        response = requests.get(f"https://api.telegram.org/bot{self.config.bot_token}/getMe")
        response.raise_for_status()



import json
from typing import Type, Any, Union

import requests
from pydantic import BaseModel, Field

from codemie_tools.base.codemie_tool import CodeMieTool
from codemie_tools.core.vcs.utils import _build_headers, file_response_handler
from .models import GithubConfig
from .tools_vars import GITHUB_TOOL

GITHUB_DEFAULT_HEADERS = {
    "Accept": "application/vnd.github+json"
}


class GithubInput(BaseModel):
    query: Union[str, dict[str, Any]] = Field(description="""
        JSON containing the GitHub API request specification. Must be valid JSON with no comments allowed.

        Required JSON structure:
        {
            "method": "GET|POST|PUT|DELETE|PATCH",
            "url": "https://api.github.com/...",
            "method_arguments": {request_parameters_or_body_data}
        }

        Optional with custom headers:
        {
            "method": "GET|POST|PUT|DELETE|PATCH",
            "url": "https://api.github.com/...", 
            "method_arguments": {request_parameters_or_body_data},
            "custom_headers": {additional_http_headers}
        }

        Field Requirements:
        - method: HTTP method (GET, POST, PUT, DELETE, PATCH) - REQUIRED
        - url: Complete GitHub API URL starting with "https://api.github.com" - REQUIRED
        - method_arguments: Object with request data (query params, body data, etc.) - REQUIRED (can be empty {})
        - custom_headers: Optional dictionary of additional HTTP headers - OPTIONAL

        Important Notes:
        - GitHub Personal Access Token is automatically added to Authorization header
        - custom_headers cannot override authorization headers (protected for security)
        - All request data goes in method_arguments regardless of HTTP method
        - Response will be raw JSON from GitHub API with automatic Base64 file decoding
        - The entire query must pass json.loads() validation

        Examples:
        Get user: {"method": "GET", "url": "https://api.github.com/user", "method_arguments": {}}
        Get repo file: {"method": "GET", "url": "https://api.github.com/repos/owner/repo/contents/file.py", "method_arguments": {}}
        Create issue: {"method": "POST", "url": "https://api.github.com/repos/owner/repo/issues", "method_arguments": {"title": "Bug", "body": "Description"}}
        """
                                              )


class GithubTool(CodeMieTool):
    name: str = GITHUB_TOOL.name
    description: str = GITHUB_TOOL.description
    args_schema: Type[BaseModel] = GithubInput
    config: GithubConfig

    # High value to support large source files.
    tokens_size_limit: int = 70_000

    @file_response_handler
    def execute(self, query: Union[str, dict[str, Any]], *args):
        """
        Execute GitHub API request with optional custom headers.

        Args:
            query: JSON containing request details

        Returns:
            JSON response from GitHub API

        Raises:
            ToolException: If credentials are missing or request fails
        """
        try:
            if isinstance(query, str):
                query = json.loads(query)
        except json.JSONDecodeError as e:
            raise ValueError(f"Query must be a JSON string: {e}")

        custom_headers = query.get('custom_headers')
        headers = _build_headers(GITHUB_DEFAULT_HEADERS, self.config.token, custom_headers)

        return requests.request(
            method=query.get('method'),
            url=query.get('url'),
            headers=headers,
            data=json.dumps(query.get('method_arguments'))
        ).json()



import io
import logging
from typing import Type, Optional, List, Dict, Any

import pandas as pd
from langchain_core.language_models import BaseChatModel
from markitdown import MarkItDown, PRIORITY_SPECIFIC_FILE_FORMAT
from pandas import DataFrame
from pydantic import BaseModel, Field

from codemie_tools.base.codemie_tool import CodeMieTool
from codemie_tools.base.constants import (
    SOURCE_DOCUMENT_KEY,
    SOURCE_FIELD_KEY,
    FILE_CONTENT_FIELD_KEY,
)
from codemie_tools.base.file_object import FileObject
from codemie_tools.base.file_tool_mixin import FileToolMixin
from codemie_tools.file_analysis.models import FileAnalysisConfig
from codemie_tools.file_analysis.tool_vars import EXCEL_TOOL
from codemie_tools.file_analysis.xlsx.markitdown_xlsx_converter import XlsxConverter
from codemie_tools.file_analysis.xlsx.processor import XlsxProcessor

logger = logging.getLogger(__name__)


class XslxToolInput(BaseModel):
    query: str = Field(
        default="", description="""User initial request should be passed as a string."""
    )
    sheet_names: List[str] = Field(
        default_factory=list,
        description="""Sheet names of Excel file to analyze. If empty, all sheets will be processed.""",
    )
    sheet_index: Optional[int] = Field(
        default=None,
        description="""Index of the sheet to analyze (0-based). If provided, overrides sheet_names.""",
    )
    get_sheet_names: bool = Field(
        default=False,
        description="""If True, returns only the names of all sheets in the Excel file.""",
    )
    get_stats: bool = Field(
        default=False,
        description="""If True, returns statistics about the Excel file (sheet count, row counts, column counts).""",
    )
    visible_only: bool = Field(
        default=True,
        description="""If True, only visible sheets will be processed. Hidden and very hidden sheets will be ignored.""",
    )
    filter_values: Optional[List[str]] = Field(
        default=None,
        description="""List of values to search for across all cells in each row. Only rows containing ALL specified values (AND logic) will be included in the output. Each value can appear in any cell of the row. Works with pivot tables and complex Excel layouts. Example: ["Critical", "Open", "Bug"] returns only rows containing all three values.""",
    )
    filter_mode: str = Field(
        default="exact",
        description="""Filter matching mode: 'exact' (default) - exact match, 'contains' - substring match. Case-insensitive. Applies to all filter values.""",
    )


class XlsxTool(CodeMieTool, FileToolMixin):
    """Tool for working with and analyzing Excel file contents."""

    args_schema: Optional[Type[BaseModel]] = XslxToolInput
    name: str = EXCEL_TOOL.name
    label: str = EXCEL_TOOL.label
    description: str = EXCEL_TOOL.description
    config: FileAnalysisConfig
    tokens_size_limit: int = 100_000

    def __init__(self, config: FileAnalysisConfig) -> None:
        """
        Initialize the XlsxTool with configuration containing Excel files.

        Args:
            config: FileAnalysisConfig with input_files and optional chat_model
        """
        super().__init__(config=config)

    def _get_supported_mime_types(self) -> Optional[List[str]]:
        """
        Get list of supported mime types for Excel processing.

        Returns:
            List of Excel mime types
        """
        return [
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
            'application/vnd.ms-excel'
        ]

    def _get_supported_extensions(self) -> Optional[List[str]]:
        """
        Get list of supported file extensions for Excel processing.

        Returns:
            List of Excel file extensions
        """
        return ['.xlsx', '.xls']

    @staticmethod
    def _load_excel_file(
        file_object: FileObject,
        clean_data: bool = True,
        visible_only: bool = True,
        sheet_names: List[str] = None,
        filter_values: Optional[List[str]] = None,
        filter_mode: str = "exact",
    ) -> Dict[str, pd.DataFrame]:
        """Load an Excel file and return a dictionary of DataFrames for each sheet

        Args:
            file_object: The FileObject containing the Excel file
            clean_data: If True, clean the data by removing empty rows and columns
            visible_only: If True, only visible sheets will be processed
            sheet_names: Optional list of specific sheet names to load
            filter_values: List of values to search for. Row must contain ALL values (AND logic).
            filter_mode: Filter matching mode ('exact' - exact match, 'contains' - substring match)

        Returns:
            Dictionary of DataFrames for each sheet
        """
        try:
            processor = XlsxProcessor(
                sheet_names=sheet_names,
                visible_only=visible_only,
                filter_values=filter_values,
                filter_mode=filter_mode,
            )
            return processor.load(file_object.bytes_content(), clean_data=clean_data)
        except Exception as e:
            logger.error(f"Failed to load Excel file: {str(e)}")
            raise e

    def _get_sheet_names(self, file_object: FileObject, visible_only: bool = True) -> List[str]:
        """Get the names of all sheets in an Excel file

        Args:
            file_object: The FileObject containing the Excel file
            visible_only: If True, only visible sheets will be returned

        Returns:
            List of sheet names
        """
        try:
            sheets = self._load_excel_file(file_object, visible_only=visible_only)
            return list(sheets.keys())
        except Exception as e:
            logger.error(f"Failed to get sheet names: {str(e)}")
            return [f"Error getting sheet names: {str(e)}"]

    def _get_sheet_by_index(
        self, file_object: FileObject, index: int, visible_only: bool = True
    ) -> tuple[DataFrame, str] | tuple[str, None]:
        """Get a specific sheet by its index (0-based)

        Args:
            file_object: The FileObject containing the Excel file
            index: The index of the sheet to get (0-based)
            visible_only: If True, only visible sheets will be considered

        Returns:
            DataFrame of the sheet and its name, or error message and None
        """
        try:
            sheets = self._load_excel_file(file_object, visible_only=visible_only)
            sheet_names = list(sheets.keys())
            if 0 <= index < len(sheet_names):
                sheet_name = sheet_names[index]
                return sheets[sheet_name], sheet_name
            else:
                return f"Invalid sheet index: {index}. Valid range: 0-{len(sheet_names)-1}", None
        except Exception as e:
            logger.error(f"Failed to get sheet by index: {str(e)}")
            return f"Error getting sheet by index: {str(e)}", None

    @staticmethod
    def _detect_column_data_type(column: pd.Series) -> str:
        """Detect the data type of a column

        Args:
            column: The pandas Series to analyze

        Returns:
            String representation of the data type
        """
        if pd.api.types.is_numeric_dtype(column):
            if pd.api.types.is_integer_dtype(column):
                return "integer"
            else:
                return "float"
        elif pd.api.types.is_datetime64_dtype(column):
            return "datetime"
        else:
            # Check if it's a boolean column (True/False values)
            unique_vals = set(column.astype(str).str.lower().unique())
            if unique_vals.issubset({"true", "false", "", "yes", "no", "y", "n"}):
                return "boolean"
            else:
                return "string"

    @staticmethod
    def _get_column_sample_values(
        column: pd.Series, max_samples: int = 5, max_length: int = 50
    ) -> List[str]:
        """Get sample values from a column

        Args:
            column: The pandas Series to get samples from
            max_samples: Maximum number of samples to return
            max_length: Maximum length of each sample string

        Returns:
            List of sample values as strings
        """
        unique_vals = column.astype(str).unique()
        # Limit samples and truncate long values
        return [
            str(val)[:max_length] + ("..." if len(str(val)) > max_length else "")
            for val in unique_vals[:max_samples]
        ]

    def _get_sheet_statistics(self, dataframe: pd.DataFrame) -> Dict[str, Any]:
        """Generate statistics for a single sheet

        Args:
            dataframe: DataFrame containing the sheet data

        Returns:
            Dictionary of sheet statistics
        """
        data_types = {}
        sample_values = {}

        for col in dataframe.columns:
            data_types[col] = self._detect_column_data_type(dataframe[col])
            sample_values[col] = self._get_column_sample_values(dataframe[col])

        return {
            "columns": list(dataframe.columns),
            "data_types": data_types,
            "sample_values": sample_values,
        }

    def _get_excel_stats(
        self, file_object: FileObject, visible_only: bool = True
    ) -> Dict[str, Any]:
        """Get comprehensive statistics about an Excel file, with data cleaning to avoid pollution

        Args:
            file_object: The FileObject containing the Excel file
            visible_only: If True, only visible sheets will be included in stats

        Returns:
            Dictionary of statistics about the Excel file
        """
        try:
            # Load cleaned sheets for clean stats
            raw_sheets = self._load_excel_file(file_object, visible_only=visible_only)

            stats = {"file_name": file_object.name, "sheet_count": len(raw_sheets), "sheets": {}}

            # Process each sheet
            for sheet_name, raw_df in raw_sheets.items():
                stats["sheets"][sheet_name] = self._get_sheet_statistics(raw_df)

            return stats
        except Exception as e:
            logger.error(f"Failed to get Excel stats: {str(e)}")
            return {"error": f"Failed to get Excel stats: {str(e)}"}

    def _process_excel_file(
        self, file_object: FileObject, sheet_names: List[str] = None, visible_only: bool = True
    ) -> str:
        """Process an Excel file and return its content as markdown text

        Args:
            file_object: The FileObject containing the Excel file
            sheet_names: List of sheet names to process. If None, all sheets will be processed
            visible_only: If True, only visible sheets will be processed

        Returns:
            Markdown text representation of the Excel file
        """
        try:
            chat_model = self.config.chat_model
            llm_model = (
                getattr(chat_model, "model_name", None)
                or getattr(chat_model, "model", None)
                if chat_model
                else None
            )
            md = MarkItDown(
                enable_builtins=True,
                llm_client=chat_model.client
                if chat_model and hasattr(chat_model, "client")
                else None,
                llm_model=llm_model,
            )
            md.register_converter(
                XlsxConverter(sheet_names=sheet_names, visible_only=visible_only),
                priority=PRIORITY_SPECIFIC_FILE_FORMAT,
            )
            # Create a file-like object from bytes content
            binary_content = io.BytesIO(file_object.bytes_content())
            result = md.convert(binary_content)
            return result.text_content
        except FileNotFoundError as e:
            # Handle the case when a file is not found
            return f"File not found: {str(e)}"
        except Exception as e:
            # Return error message for Excel processing failures
            return f"Failed to process Excel file: {str(e)}"

    @staticmethod
    def _format_dataframe_as_markdown(df: pd.DataFrame, sheet_name: str = None) -> str:
        """Format a DataFrame as a markdown table"""
        header = f"## {sheet_name}\n" if sheet_name else ""
        return header + df.to_markdown(index=False)

    @staticmethod
    def _format_stats_as_markdown(stats: Dict[str, Any]) -> str:
        """Format Excel statistics as markdown with comprehensive details"""
        if "error" in stats:
            return f"Error: {stats['error']}"

        results = [
            f"# Excel File Statistics: {stats['file_name']}",
            f"- **Total Sheets:** {stats['sheet_count']}",
        ]

        for sheet_name, sheet_stats in stats["sheets"].items():
            results.append(f"\n## Sheet: {sheet_name}")

            # Column information
            if "columns" in sheet_stats:
                results.append("\n### Columns:")
                results.append("| Column | Data Type | Sample Values |")
                results.append("| ------ | --------- | ------------- |")

                for col in sheet_stats["columns"]:
                    data_type = sheet_stats.get("data_types", {}).get(col, "unknown")
                    samples = sheet_stats.get("sample_values", {}).get(col, [])
                    sample_str = ", ".join([f"`{s}`" for s in samples[:3]])
                    if len(samples) > 3:
                        sample_str += ", ..."
                    results.append(f"| {col} | {data_type} | {sample_str} |")

        return "\n".join(results)

    def execute(self, query: str = "", **kwargs):
        """Execute the Excel analysis tool with various options.

        Args:
            query: User's initial request string
            **kwargs: Additional parameters including sheet_names, sheet_index, etc.

        Returns:
            Processed Excel content as markdown string
        """
        # Get supported Excel files from config (automatically filters by mime type)
        files = self._get_supported_files()

        if not files:
            raise ValueError(f"{self.name} requires at least one Excel file to process.")

        # Extract parameters from kwargs with defaults
        sheet_names = kwargs.get("sheet_names")
        sheet_index = kwargs.get("sheet_index")
        get_sheet_names = kwargs.get("get_sheet_names", False)
        get_stats = kwargs.get("get_stats", False)
        visible_only = kwargs.get("visible_only", True)
        filter_values = kwargs.get("filter_values")
        filter_mode = kwargs.get("filter_mode", "exact")

        return self._process_files(
            sheet_names,
            sheet_index,
            get_sheet_names,
            get_stats,
            visible_only,
            filter_values,
            filter_mode,
        )

    def _process_files(
        self,
        sheet_names: List[str],
        sheet_index: int,
        get_sheet_names: bool,
        get_stats: bool,
        visible_only: bool,
        filter_values: Optional[List[str]],
        filter_mode: str,
    ) -> str:
        """Process all files with the specified parameters

        Returns:
            Joined results from all files
        """
        files = self._get_supported_files()

        # Process multiple Excel files with LLM-friendly separators
        result = []

        for file_object in files:
            content = self._process_file_content(
                file_object,
                sheet_names,
                sheet_index,
                get_sheet_names,
                get_stats,
                visible_only,
                filter_values,
                filter_mode,
            )

            # Add file header with metadata
            result.append(f"\n{SOURCE_DOCUMENT_KEY}\n")
            result.append(f"{SOURCE_FIELD_KEY} {file_object.name}\n")
            result.append(f"{FILE_CONTENT_FIELD_KEY} \n{content}\n")

        return "\n".join(result)

    def _process_file_content(
        self,
        file_object: FileObject,
        sheet_names: List[str],
        sheet_index: int,
        get_sheet_names: bool,
        get_stats: bool,
        visible_only: bool,
        filter_values: Optional[List[str]],
        filter_mode: str,
    ) -> str:
        """Process file content based on various options

        Args:
            file_object: The FileObject containing Excel file
            sheet_names: List of sheet names
            sheet_index: Specific sheet index
            get_sheet_names: Whether to only get sheet names
            get_stats: Whether to get statistics
            visible_only: Whether to only process visible sheets
            filter_values: List of values to filter rows (AND logic)
            filter_mode: Filter mode (exact or contains)

        Returns:
            Processed content as string
        """
        # Handle special operations
        if get_sheet_names:
            return self._get_sheet_names_content(file_object, visible_only)

        if get_stats:
            return self._get_stats_content(file_object, visible_only)

        if sheet_index is not None:
            return self._get_sheet_by_index_content(
                file_object, sheet_index, visible_only, filter_values, filter_mode
            )

        # Check if we need filtering
        if filter_values:
            return self._get_filtered_content(
                file_object, sheet_names, visible_only, filter_values, filter_mode
            )

        # Default: standard processing
        return self._process_excel_file(
            file_object, sheet_names=sheet_names, visible_only=visible_only
        )

    def _get_sheet_names_content(self, file_object: FileObject, visible_only: bool) -> str:
        """Get sheet names as formatted content"""
        sheet_list = self._get_sheet_names(file_object, visible_only=visible_only)
        return f"## Sheets in {file_object.name}:\n- " + "\n- ".join(sheet_list)

    def _get_stats_content(self, file_object: FileObject, visible_only: bool) -> str:
        """Get statistics as formatted content"""
        stats = self._get_excel_stats(file_object, visible_only=visible_only)
        return self._format_stats_as_markdown(stats)

    def _get_filtered_content(
        self,
        file_object: FileObject,
        sheet_names: List[str],
        visible_only: bool,
        filter_values: List[str],
        filter_mode: str,
    ) -> str:
        """Get filtered content as markdown"""
        sheets = self._load_excel_file(
            file_object,
            sheet_names=sheet_names,
            visible_only=visible_only,
            filter_values=filter_values,
            filter_mode=filter_mode,
        )
        return self._format_sheets_as_markdown(sheets)

    def _format_sheets_as_markdown(self, sheets: Dict[str, pd.DataFrame]) -> str:
        """Format multiple sheets as markdown tables

        Args:
            sheets: Dictionary of sheet_name -> DataFrame

        Returns:
            Formatted markdown string with all sheets
        """
        content_parts = []
        for sheet_name, df in sheets.items():
            content_parts.append(self._format_dataframe_as_markdown(df, sheet_name))
        return "\n\n".join(content_parts)

    def _get_sheet_by_index_content(
        self,
        file_object: FileObject,
        sheet_index: int,
        visible_only: bool,
        filter_values: Optional[List[str]],
        filter_mode: str,
    ) -> str:
        """Get content for a specific sheet by index

        Args:
            file_object: The FileObject containing Excel file
            sheet_index: Sheet index to retrieve
            visible_only: Whether to only consider visible sheets
            filter_values: List of values to filter rows (AND logic)
            filter_mode: Filter mode (exact or contains)

        Returns:
            Sheet content as string
        """
        sheets = self._load_excel_file(
            file_object,
            visible_only=visible_only,
            filter_values=filter_values,
            filter_mode=filter_mode,
        )
        sheet_names_list = list(sheets.keys())

        if 0 <= sheet_index < len(sheet_names_list):
            sheet_name = sheet_names_list[sheet_index]
            sheet_data = sheets[sheet_name]
            return self._format_dataframe_as_markdown(sheet_data, sheet_name)

        return f"Invalid sheet index: {sheet_index}. Valid range: 0-{len(sheet_names_list)-1}"



"""Xray Cloud tools for test management."""

import json
from typing import Dict, List, Optional, Type

from langchain_core.tools import ToolException
from pydantic import BaseModel

from codemie_tools.base.codemie_tool import CodeMieTool
from codemie_tools.qa.xray.models import (
    XrayConfig,
    XrayGetTestsInput,
    XrayCreateTestInput,
    XrayExecuteGraphQLInput
)
from codemie_tools.qa.xray.tools_vars import (
    XRAY_GET_TESTS_TOOL,
    XRAY_CREATE_TEST_TOOL,
    XRAY_EXECUTE_GRAPHQL_TOOL
)
from codemie_tools.qa.xray.xray_client import XrayClient

# Error message constants
ERROR_CONFIG_NOT_PROVIDED = "Xray config is not provided"
ERROR_CONFIG_NOT_SET = "Xray config is not provided. Please set it before using the tool."


class XrayGetTestsTool(CodeMieTool):
    """Tool for retrieving test cases from Xray Cloud using JQL queries."""

    config: Optional[XrayConfig] = None
    name: str = XRAY_GET_TESTS_TOOL.name
    description: str = XRAY_GET_TESTS_TOOL.description
    args_schema: Type[BaseModel] = XrayGetTestsInput

    def _healthcheck(self):
        """Perform health check by verifying authentication."""
        if not self.config:
            raise ToolException(ERROR_CONFIG_NOT_PROVIDED)

        client = self._create_client()
        client.health_check()

    def _create_client(self) -> XrayClient:
        """Create and return XrayClient instance."""
        return XrayClient(
            base_url=self.config.base_url,
            client_id=self.config.client_id,
            client_secret=self.config.client_secret,
            limit=self.config.limit,
            verify_ssl=self.config.verify_ssl
        )

    def _format_test_item(self, test: Dict) -> List[str]:
        """Format a single test item for display."""
        lines = []
        jira_info = test.get("jira", {})
        test_type = test.get("testType", {}).get("name", "Unknown")
        issue_key = jira_info.get("key", "N/A") if isinstance(jira_info, dict) else jira_info
        summary = jira_info.get("summary", "") if isinstance(jira_info, dict) else ""

        lines.append(f"  - [{issue_key}] {summary} (Type: {test_type})")

        # Include step count for manual tests
        if test_type == "Manual" and "steps" in test:
            step_count = len(test["steps"])
            lines.append(f"    Steps: {step_count}")

        return lines

    def execute(self, jql: str) -> str:
        """
        Retrieve test cases from Xray using JQL query.

        Args:
            jql: JQL query string to filter tests

        Returns:
            str: Formatted string with test count and test details

        Raises:
            ToolException: If config is missing or query execution fails
        """
        if not self.config:
            raise ToolException(ERROR_CONFIG_NOT_SET)

        try:
            client = self._create_client()
            result = client.get_tests(jql)

            total = result.get("total", 0)
            tests = result.get("tests", [])

            # Format the response
            response_lines = [f"Retrieved {total} test(s) matching query: {jql}"]

            if tests:
                response_lines.append("\nTests:")
                for test in tests:
                    response_lines.extend(self._format_test_item(test))

                # Add full test details as JSON for reference
                response_lines.append("\nFull test details:")
                response_lines.append(json.dumps(tests, indent=2))
            else:
                response_lines.append("\nNo tests found matching the query.")

            return "\n".join(response_lines)

        except Exception as e:
            raise ToolException(f"Failed to retrieve tests: {str(e)}")


class XrayCreateTestTool(CodeMieTool):
    """Tool for creating new test cases in Xray Cloud using GraphQL mutations."""

    config: Optional[XrayConfig] = None
    name: str = XRAY_CREATE_TEST_TOOL.name
    description: str = XRAY_CREATE_TEST_TOOL.description
    args_schema: Type[BaseModel] = XrayCreateTestInput

    def _healthcheck(self):
        """Perform health check by verifying authentication."""
        if not self.config:
            raise ToolException(ERROR_CONFIG_NOT_PROVIDED)

        client = self._create_client()
        client.health_check()

    def _create_client(self) -> XrayClient:
        """Create and return XrayClient instance."""
        return XrayClient(
            base_url=self.config.base_url,
            client_id=self.config.client_id,
            client_secret=self.config.client_secret,
            limit=self.config.limit,
            verify_ssl=self.config.verify_ssl
        )

    def _format_test_details(self, test: Dict) -> List[str]:
        """Format test details for display."""
        lines = []
        jira_info = test.get("jira", {})
        test_type = test.get("testType", {}).get("name", "Unknown")
        issue_id = test.get("issueId", "N/A")
        issue_key = jira_info.get("key", "N/A") if isinstance(jira_info, dict) else jira_info

        lines.append("\nTest Details:")
        lines.append(f"  - Issue ID: {issue_id}")
        lines.append(f"  - Issue Key: {issue_key}")
        lines.append(f"  - Test Type: {test_type}")

        # Add type-specific details
        if "steps" in test and test["steps"]:
            lines.append(f"  - Steps: {len(test['steps'])}")
        elif "unstructured" in test:
            lines.append(f"  - Description: {test['unstructured'][:100]}...")
        elif "gherkin" in test:
            lines.append("  - Type: Cucumber (Gherkin)")

        return lines

    def execute(self, graphql_mutation: str) -> str:
        """
        Create a new test case in Xray using GraphQL mutation.

        Args:
            graphql_mutation: GraphQL mutation string to create test

        Returns:
            str: Formatted string with created test details

        Raises:
            ToolException: If config is missing or test creation fails
        """
        if not self.config:
            raise ToolException(ERROR_CONFIG_NOT_SET)

        try:
            client = self._create_client()
            result = client.create_test(graphql_mutation)

            test = result.get("test", {})
            warnings = result.get("warnings", [])

            # Format the response
            response_lines = ["Test created successfully!"]

            if test:
                response_lines.extend(self._format_test_details(test))

            if warnings:
                response_lines.append("\nWarnings:")
                for warning in warnings:
                    response_lines.append(f"  - {warning}")

            # Add full response as JSON for reference
            response_lines.append("\nFull response:")
            response_lines.append(json.dumps(result, indent=2))

            return "\n".join(response_lines)

        except Exception as e:
            raise ToolException(f"Failed to create test: {str(e)}")


class XrayExecuteGraphQLTool(CodeMieTool):
    """Tool for executing custom GraphQL queries and mutations against Xray Cloud API."""

    config: Optional[XrayConfig] = None
    name: str = XRAY_EXECUTE_GRAPHQL_TOOL.name
    description: str = XRAY_EXECUTE_GRAPHQL_TOOL.description
    args_schema: Type[BaseModel] = XrayExecuteGraphQLInput

    def _healthcheck(self):
        """Perform health check by verifying authentication."""
        if not self.config:
            raise ToolException(ERROR_CONFIG_NOT_PROVIDED)

        client = self._create_client()
        client.health_check()

    def _create_client(self) -> XrayClient:
        """Create and return XrayClient instance."""
        return XrayClient(
            base_url=self.config.base_url,
            client_id=self.config.client_id,
            client_secret=self.config.client_secret,
            limit=self.config.limit,
            verify_ssl=self.config.verify_ssl
        )

    def execute(self, graphql: str) -> str:
        """
        Execute custom GraphQL query or mutation.

        Args:
            graphql: Custom GraphQL query or mutation string

        Returns:
            str: Formatted string with GraphQL execution result

        Raises:
            ToolException: If config is missing or execution fails
        """
        if not self.config:
            raise ToolException(ERROR_CONFIG_NOT_SET)

        try:
            client = self._create_client()
            result = client.execute_custom_graphql(graphql)

            # Format the response
            response_lines = ["GraphQL executed successfully!"]
            response_lines.append("\nResult:")
            response_lines.append(json.dumps(result, indent=2))

            return "\n".join(response_lines)

        except Exception as e:
            raise ToolException(f"Failed to execute GraphQL: {str(e)}")



"""Cloud toolkit for AWS, Azure, GCP, and Kubernetes integrations."""

from .aws.tools import GenericAWSTool
from .azure.tools import GenericAzureTool
from .gcp.tools import GenericGCPTool
from .kubernetes.tools import GenericKubernetesTool
from .toolkit import CloudToolkit, CloudToolkitUI

__all__ = [
    "CloudToolkit",
    "CloudToolkitUI",
    "GenericAWSTool",
    "GenericAzureTool",
    "GenericGCPTool",
    "GenericKubernetesTool",
]



from typing import Any, Optional, List

from pydantic import Field

from codemie_tools.base.base_toolkit import BaseToolkit
from codemie_tools.base.file_object import FileObject
from codemie_tools.base.models import ToolKit, ToolSet, Tool
from codemie_tools.vision.image_tool import ImageTool
from codemie_tools.vision.tool_vars import IMAGE_TOOL


class VisionToolkit(BaseToolkit):
    chat_model: Optional[Any] = None
    files: list[FileObject] = Field(default_factory=list, exclude=True)

    @classmethod
    def get_tools_ui_info(cls, *args, **kwargs):
        return ToolKit(
            toolkit=ToolSet.VISION.value,
            tools=[
                Tool.from_metadata(IMAGE_TOOL),
            ]
        ).model_dump()

    def get_tools(self):
        """Get tools for processing images.
    
        Returns:
            List of tools for image processing
        """
        tools = []

        # Process files if available
        if not self.files:
            return tools

        tools.append(ImageTool(files=self.files, chat_model=self.chat_model))

        return tools

    @classmethod
    def get_toolkit(
            cls,
            files: List[FileObject],
            chat_model: Optional[Any] = None
    ):
        # Filter for image files
        image_files = []
        for file_obj in files:
            if file_obj.is_image():
                image_files.append(file_obj)
        return VisionToolkit(files=image_files, chat_model=chat_model)



import json
import os
from typing import Type, Optional

from azure.devops.connection import Connection
from azure.devops.v7_0.test_plan.models import (
    TestPlanCreateParams,
    TestSuiteCreateParams,
    SuiteTestCaseCreateUpdateParameters,
)
from azure.devops.v7_0.test_plan.test_plan_client import TestPlanClient
from langchain_core.tools import ToolException
from msrest.authentication import BasicAuthentication
from pydantic import BaseModel

from codemie_tools.azure_devops.test_plan.models import (
    AzureDevOpsTestPlanConfig,
    CreateTestPlanInput,
    DeleteTestPlanInput,
    GetTestPlanInput,
    CreateTestSuiteInput,
    DeleteTestSuiteInput,
    GetTestSuiteInput,
    AddTestCaseInput,
    GetTestCaseInput,
    GetTestCasesInput,
)
from codemie_tools.azure_devops.test_plan.tools_vars import (
    CREATE_TEST_PLAN_TOOL,
    DELETE_TEST_PLAN_TOOL,
    GET_TEST_PLAN_TOOL,
    CREATE_TEST_SUITE_TOOL,
    DELETE_TEST_SUITE_TOOL,
    GET_TEST_SUITE_TOOL,
    ADD_TEST_CASE_TOOL,
    GET_TEST_CASE_TOOL,
    GET_TEST_CASES_TOOL,
)
from codemie_tools.base.codemie_tool import CodeMieTool, logger

# Ensure Azure DevOps cache directory is set
if not os.environ.get("AZURE_DEVOPS_CACHE_DIR", None):
    os.environ["AZURE_DEVOPS_CACHE_DIR"] = ""


class BaseAzureDevOpsTestPlanTool(CodeMieTool):
    """Base class for Azure DevOps Test Plan tools."""

    config: AzureDevOpsTestPlanConfig
    __client: Optional[TestPlanClient] = None

    @property
    def _client(self) -> TestPlanClient:
        """Get or create Azure DevOps client (lazy initialization)."""
        if self.__client is None:
            try:
                # Set up connection to Azure DevOps using Personal Access Token (PAT)
                credentials = BasicAuthentication("", self.config.token)
                connection = Connection(base_url=self.config.organization_url, creds=credentials)
                # Retrieve the test plan client
                self.__client = connection.clients.get_test_plan_client()
            except Exception as e:
                logger.error(f"Failed to connect to Azure DevOps: {e}")
                raise ToolException(f"Failed to connect to Azure DevOps: {e}")
        return self.__client

    @_client.setter
    def _client(self, value: TestPlanClient) -> None:
        """Set the Azure DevOps client (useful for testing)."""
        self.__client = value


class CreateTestPlanTool(BaseAzureDevOpsTestPlanTool):
    """Tool to create a test plan in Azure DevOps."""

    name: str = CREATE_TEST_PLAN_TOOL.name
    description: str = CREATE_TEST_PLAN_TOOL.description
    args_schema: Type[BaseModel] = CreateTestPlanInput

    def execute(self, test_plan_create_params: str, project: Optional[str] = None):
        """Create a test plan in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            params = json.loads(test_plan_create_params)
            test_plan_create_params_obj = TestPlanCreateParams(**params)
            test_plan = self._client.create_test_plan(test_plan_create_params_obj, project_to_use)
            return f"Test plan {test_plan.id} created successfully."
        except Exception as e:
            logger.error(f"Error creating test plan: {e}")
            raise ToolException(f"Error creating test plan: {e}")


class DeleteTestPlanTool(BaseAzureDevOpsTestPlanTool):
    """Tool to delete a test plan in Azure DevOps."""

    name: str = DELETE_TEST_PLAN_TOOL.name
    description: str = DELETE_TEST_PLAN_TOOL.description
    args_schema: Type[BaseModel] = DeleteTestPlanInput

    def execute(self, plan_id: int, project: Optional[str] = None):
        """Delete a test plan in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            self._client.delete_test_plan(project_to_use, plan_id)
            return f"Test plan {plan_id} deleted successfully."
        except Exception as e:
            logger.error(f"Error deleting test plan: {e}")
            raise ToolException(f"Error deleting test plan: {e}")


class GetTestPlanTool(BaseAzureDevOpsTestPlanTool):
    """Tool to get a test plan or list test plans in Azure DevOps."""

    name: str = GET_TEST_PLAN_TOOL.name
    description: str = GET_TEST_PLAN_TOOL.description
    args_schema: Type[BaseModel] = GetTestPlanInput

    def execute(self, plan_id: Optional[int] = None, project: Optional[str] = None):
        """Get a test plan or list of test plans in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            if plan_id:
                test_plan = self._client.get_test_plan_by_id(project_to_use, plan_id)
                return test_plan.as_dict()
            else:
                test_plans = self._client.get_test_plans(project_to_use)
                return [plan.as_dict() for plan in test_plans[: self.config.limit]]
        except Exception as e:
            logger.error(f"Error getting test plan(s): {e}")
            raise ToolException(f"Error getting test plan(s): {e}")


class CreateTestSuiteTool(BaseAzureDevOpsTestPlanTool):
    """Tool to create a test suite in Azure DevOps."""

    name: str = CREATE_TEST_SUITE_TOOL.name
    description: str = CREATE_TEST_SUITE_TOOL.description
    args_schema: Type[BaseModel] = CreateTestSuiteInput

    def execute(self, test_suite_create_params: str, plan_id: int, project: Optional[str] = None):
        """Create a test suite in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            params = json.loads(test_suite_create_params)
            test_suite_create_params_obj = TestSuiteCreateParams(**params)
            test_suite = self._client.create_test_suite(
                test_suite_create_params_obj, project_to_use, plan_id
            )
            return f"Test suite {test_suite.id} created successfully."
        except Exception as e:
            logger.error(f"Error creating test suite: {e}")
            raise ToolException(f"Error creating test suite: {e}")


class DeleteTestSuiteTool(BaseAzureDevOpsTestPlanTool):
    """Tool to delete a test suite in Azure DevOps."""

    name: str = DELETE_TEST_SUITE_TOOL.name
    description: str = DELETE_TEST_SUITE_TOOL.description
    args_schema: Type[BaseModel] = DeleteTestSuiteInput

    def execute(self, plan_id: int, suite_id: int, project: Optional[str] = None):
        """Delete a test suite in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            self._client.delete_test_suite(project_to_use, plan_id, suite_id)
            return f"Test suite {suite_id} deleted successfully."
        except Exception as e:
            logger.error(f"Error deleting test suite: {e}")
            raise ToolException(f"Error deleting test suite: {e}")


class GetTestSuiteTool(BaseAzureDevOpsTestPlanTool):
    """Tool to get a test suite or list test suites in Azure DevOps."""

    name: str = GET_TEST_SUITE_TOOL.name
    description: str = GET_TEST_SUITE_TOOL.description
    args_schema: Type[BaseModel] = GetTestSuiteInput

    def execute(self, plan_id: int, suite_id: Optional[int] = None, project: Optional[str] = None):
        """Get a test suite or list of test suites in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            if suite_id:
                test_suite = self._client.get_test_suite_by_id(project_to_use, plan_id, suite_id)
                return test_suite.as_dict()
            else:
                test_suites = self._client.get_test_suites_for_plan(project_to_use, plan_id)
                return [suite.as_dict() for suite in test_suites[: self.config.limit]]
        except Exception as e:
            logger.error(f"Error getting test suite(s): {e}")
            raise ToolException(f"Error getting test suite(s): {e}")


class AddTestCaseTool(BaseAzureDevOpsTestPlanTool):
    """Tool to add a test case to a suite in Azure DevOps."""

    name: str = ADD_TEST_CASE_TOOL.name
    description: str = ADD_TEST_CASE_TOOL.description
    args_schema: Type[BaseModel] = AddTestCaseInput

    def execute(
        self,
        suite_test_case_create_update_parameters: str,
        plan_id: int,
        suite_id: int,
        project: Optional[str] = None,
    ):
        """Add a test case to a suite in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            params = json.loads(suite_test_case_create_update_parameters)
            suite_test_case_params_list = []

            # Handle both array and single object scenarios
            if isinstance(params, list):
                for param in params:
                    suite_test_case_params_list.append(SuiteTestCaseCreateUpdateParameters(**param))
            else:
                suite_test_case_params_list.append(SuiteTestCaseCreateUpdateParameters(**params))

            test_cases = self._client.add_test_cases_to_suite(
                suite_test_case_params_list, project_to_use, plan_id, suite_id
            )
            return [test_case.as_dict() for test_case in test_cases]
        except Exception as e:
            logger.error(f"Error adding test case: {e}")
            raise ToolException(f"Error adding test case: {e}")


class GetTestCaseTool(BaseAzureDevOpsTestPlanTool):
    """Tool to get a test case from a suite in Azure DevOps."""

    name: str = GET_TEST_CASE_TOOL.name
    description: str = GET_TEST_CASE_TOOL.description
    args_schema: Type[BaseModel] = GetTestCaseInput

    def execute(
        self, plan_id: int, suite_id: int, test_case_id: str, project: Optional[str] = None
    ):
        """Get a test case from a suite in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            test_cases = self._client.get_test_case(project_to_use, plan_id, suite_id, test_case_id)
            if test_cases:  # Check if the list is not empty
                test_case = test_cases[0]
                return test_case.as_dict()
            else:
                return f"No test cases found with criteria: project {project_to_use}, plan {plan_id}, suite {suite_id}, test case id {test_case_id}"
        except Exception as e:
            logger.error(f"Error getting test case: {e}")
            raise ToolException(f"Error getting test case: {e}")


class GetTestCasesTool(BaseAzureDevOpsTestPlanTool):
    """Tool to get test cases from a suite in Azure DevOps."""

    name: str = GET_TEST_CASES_TOOL.name
    description: str = GET_TEST_CASES_TOOL.description
    args_schema: Type[BaseModel] = GetTestCasesInput

    def execute(self, plan_id: int, suite_id: int, project: Optional[str] = None):
        """Get test cases from a suite in Azure DevOps."""
        # Use provided project or default from config if not provided
        project_to_use = project or self.config.project

        try:
            test_cases = self._client.get_test_case_list(project_to_use, plan_id, suite_id)
            return [test_case.as_dict() for test_case in test_cases[: self.config.limit]]
        except Exception as e:
            logger.error(f"Error getting test cases: {e}")
            raise ToolException(f"Error getting test cases: {e}")



import base64
import logging
from typing import Optional

import cv2
import numpy as np
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import HumanMessage

# Configure logger
logger = logging.getLogger(__name__)

# Constants for LLM-based image text extraction
DEFAULT_IMAGE_TEXT_PROMPT = """
Extract all text visible in this image. Format it properly preserving paragraphs, 
bullet points, and other structural elements.
"""

MAX_TOKENS = 5000


class ImageProcessor:
    """
    Utility class for processing images and extracting text using LLM-based recognition.
    """

    def __init__(self, chat_model: Optional[BaseChatModel] = None):
        """
        Initialize the ImageProcessor.

        Args:
            chat_model (Optional[BaseChatModel]): The language model to use for image text extraction.
        """
        self.chat_model = chat_model

    def encode_image_base64(self, image_bytes: bytes) -> str:
        """
        Encode image bytes to base64 with proper data URL prefix.
        
        Args:
            image_bytes: Raw image bytes to encode
            
        Returns:
            str: Base64 encoded image with data URL prefix
        """
        if not image_bytes:
            return ""
        
        encoded = base64.b64encode(image_bytes).decode('utf-8')
        return f"data:image/jpeg;base64,{encoded}"

    def extract_text_from_image_bytes(self, image_bytes: bytes, custom_prompt: Optional[str] = None) -> str:
        """
        Extract text from image bytes using LLM vision capabilities.

        Args:
            image_bytes (bytes): Raw image bytes to process
            custom_prompt (Optional[str]): Custom prompt to use for the vision model

        Returns:
            str: Extracted text from the image
        """
        try:
            # Convert to numpy array for OpenCV
            nparr = np.frombuffer(image_bytes, np.uint8)
            image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)

            if image is None:
                logger.warning("Failed to decode image bytes")
                return ""

            # Convert OpenCV image to JPEG bytes and encode as base64
            _, buffer = cv2.imencode('.jpg', image)
            jpg_as_text = base64.b64encode(buffer).decode('utf-8')

            return self._extract_text_using_llm(jpg_as_text, custom_prompt)

        except Exception as e:
            logger.error(f"Error processing image bytes: {str(e)}")
            return ""

    def _extract_text_using_llm(self, base64_image: str, custom_prompt: Optional[str] = None) -> str:
        """
        Extract text from a base64-encoded image using LLM vision capabilities.

        Args:
            base64_image (str): The base64-encoded image data
            custom_prompt (Optional[str]): Custom prompt to use instead of default

        Returns:
            str: The extracted text from the image
        """
        if not self.chat_model:
            logger.warning("No chat model provided for image text extraction")
            return ""

        prompt = custom_prompt if custom_prompt is not None else DEFAULT_IMAGE_TEXT_PROMPT
        
        try:
            # Create the message with image data
            image_message = HumanMessage(
                content=[
                    {"type": "text", "text": prompt},
                    {"type": "image_url",
                     "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"}}
                ]
            )

            # Get text from the chat model
            response = self.chat_model.invoke(
                [image_message],
                max_tokens=MAX_TOKENS
            )

            return response.content

        except Exception as e:
            logger.error(f"Error extracting text from image using LLM: {str(e)}")
            return ""


from typing import Optional

from pydantic import BaseModel, Field


class ListBranchesToolInput(BaseModel):
    query: Optional[str] = Field(
        default="",
        description="User initial request should be passed as a string.",
    )



from typing import Dict, Any, Optional
from urllib.parse import urlparse

import httpx
from azure.identity import ClientSecretCredential
from langchain_core.tools import ToolException


class AzureClient:
    """HTTP client for Azure REST API."""

    # Whitelisted Azure domains
    WHITE_DOMAINS = [
        ".azure.com",
        ".azure.net",
        ".microsoft.com",
        ".windows.net",
        ".microsoftonline.com"
    ]

    def __init__(
        self,
        subscription_id: str,
        tenant_id: str,
        client_id: str,
        client_secret: str,
        timeout: int = 30
    ):
        """
        Initialize the Azure client.

        Args:
            subscription_id: Azure subscription ID
            tenant_id: Azure tenant ID
            client_id: Azure client (application) ID
            client_secret: Azure client secret
            timeout: Request timeout in seconds
        """
        self.subscription_id = subscription_id
        self.tenant_id = tenant_id
        self.timeout = timeout
        self.client_id = client_id
        self.client_secret = client_secret

    def get_credentials(self) -> ClientSecretCredential:
        # Initialize Azure credential
        try:
            return ClientSecretCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
        except Exception as e:
            raise ToolException(f"Failed to initialize Azure credentials: {str(e)}")

    def _validate_domain(self, url: str) -> None:
        """
        Validate that the URL domain is whitelisted for Azure.

        Args:
            url: The URL to validate

        Raises:
            ToolException: If domain is not whitelisted
        """
        domain = urlparse(url).netloc
        if not any(domain.endswith(allowed_domain) for allowed_domain in self.WHITE_DOMAINS):
            raise ToolException(
                f"Domain '{domain}' is not whitelisted for Azure requests. "
                f"Allowed domains: {', '.join(self.WHITE_DOMAINS)}"
            )

    def _get_access_token(self, scope: str) -> str:
        """
        Get an access token for the specified scope.

        Args:
            scope: OAuth scope for the token

        Returns:
            str: Access token

        Raises:
            ToolException: If token acquisition fails
        """
        try:
            token = self.get_credentials().get_token(scope)
            return token.token
        except Exception as e:
            raise ToolException(f"Failed to acquire Azure access token: {str(e)}")

    def request(
        self,
        method: str,
        url: str,
        scope: str = "https://management.azure.com/.default",
        optional_args: Optional[Dict[str, Any]] = None
    ) -> str:
        """
        Make an HTTP request to Azure REST API.

        Args:
            method: HTTP method (GET, POST, PUT, DELETE, etc.)
            url: Full URL for the request
            scope: OAuth scope for authentication
            optional_args: Optional request parameters (data, headers, params, etc.)

        Returns:
            str: Response text

        Raises:
            ToolException: If request fails
        """
        # Validate URL domain
        self._validate_domain(url)

        # Get access token
        token = self._get_access_token(scope)

        # Prepare headers
        headers = {"Authorization": f"Bearer {token}"}

        # Merge optional headers if provided
        request_kwargs = optional_args or {}
        if "headers" in request_kwargs:
            request_kwargs["headers"].update(headers)
        else:
            request_kwargs["headers"] = headers

        # Make the request
        try:
            with httpx.Client(timeout=self.timeout) as client:
                response = client.request(
                    method=method.upper(),
                    url=url,
                    **request_kwargs
                )

                # Check status code
                if response.status_code >= 400:
                    raise ToolException(
                        f"Azure API error: HTTP {response.status_code} - {response.text}"
                    )

                return response.text

        except httpx.HTTPStatusError as e:
            raise ToolException(
                f"HTTP {e.response.status_code}: {e.response.text}"
            )
        except httpx.RequestError as e:
            raise ToolException(f"Request failed: {str(e)}")
        except ToolException:
            raise
        except Exception as e:
            raise ToolException(f"Unexpected error: {str(e)}")

    def health_check(self):
        """
        Check if Azure service is accessible by listing resource groups.
        """
        if not self.subscription_id:
            raise ToolException("Subscription ID is required.")
        url = f"https://management.azure.com/subscriptions/{self.subscription_id}/resourcegroups?api-version=2021-04-01"
        self.request("GET", url)



"""
Pod discovery service for Kubernetes-based code execution.

This module provides services for discovering and managing executor pods
in a Kubernetes cluster, separated from session management for clarity.
"""

import logging
import time
from typing import List, Optional

logger = logging.getLogger(__name__)


class PodDiscoveryService:
    """
    Service for discovering and validating Kubernetes pods.

    Handles all pod-related queries and validation, keeping pod discovery
    logic separate from session management.
    """

    def __init__(self, client_manager, namespace: str):
        """
        Initialize pod discovery service.

        Args:
            client_manager: KubernetesClientManager for getting/recreating client
            namespace: Kubernetes namespace to search for pods
        """
        self.client_manager = client_manager
        self.namespace = namespace

    @property
    def k8s_client(self):
        """Get Kubernetes client from manager."""
        return self.client_manager.get_client()

    def list_available_pods(self) -> List[str]:
        """
        List all running pods with ready containers in the namespace.

        Only returns pods where:
        1. Pod phase is "Running"
        2. All containers have ready=True status

        Automatically recreates Kubernetes client if connection pool corruption is detected.

        Returns:
            List[str]: Names of available ready pods
        """
        try:
            pods = self.k8s_client.list_namespaced_pod(
                namespace=self.namespace,
                label_selector="app=codemie-executor",
            )
            return self._filter_ready_pods(pods)

        except Exception as e:
            # Check if this is a connection pool corruption error
            if self._is_connection_error(str(e)):
                logger.debug(
                    f"Kubernetes client connection pool corrupted ({type(e).__name__}), "
                    "will retry with fresh client"
                )
                # Recreate client and retry once
                self.client_manager.recreate_client()
                try:
                    pods = self.k8s_client.list_namespaced_pod(
                        namespace=self.namespace,
                        label_selector="app=codemie-executor",
                    )
                    return self._filter_ready_pods(pods)
                except Exception as retry_error:
                    logger.warning(f"Retry after client recreation also failed: {retry_error}")
                    return []

            logger.warning(f"Failed to list pods: {e}")
            return []

    def check_pod_exists_and_ready(self, pod_name: str) -> bool:
        """
        Check if a pod exists, is Running, and all containers are ready.

        Automatically recreates Kubernetes client if connection pool corruption is detected.

        Args:
            pod_name: Name of the pod to check

        Returns:
            bool: True if pod exists, is running, and containers are ready
        """
        try:
            pod = self.k8s_client.read_namespaced_pod(
                name=pod_name, namespace=self.namespace
            )

            if pod.status.phase != "Running":
                logger.debug(f"Pod {pod_name} not running (phase: {pod.status.phase})")
                return False

            return self._check_containers_ready(pod, pod_name)

        except Exception as e:
            # Check if this is a connection pool corruption error
            if self._is_connection_error(str(e)):
                logger.debug(
                    f"Kubernetes client connection pool corrupted ({type(e).__name__}), "
                    "will retry with fresh client"
                )
                # Recreate client and retry once
                self.client_manager.recreate_client()
                try:
                    pod = self.k8s_client.read_namespaced_pod(
                        name=pod_name, namespace=self.namespace
                    )

                    if pod.status.phase != "Running":
                        logger.debug(f"Pod {pod_name} not running (phase: {pod.status.phase})")
                        return False

                    return self._check_containers_ready(pod, pod_name)
                except Exception as retry_error:
                    logger.debug(f"Pod {pod_name} check failed after retry: {retry_error}")
                    return False

            logger.debug(f"Pod {pod_name} does not exist or cannot be accessed: {e}")
            return False

    def wait_for_available_pods(self, max_retries: int = 20, retry_delay: float = 0.5) -> List[str]:
        """
        Poll for available Running pods with retry logic.

        Args:
            max_retries: Maximum number of retry attempts
            retry_delay: Delay between retries in seconds

        Returns:
            List[str]: List of available pod names (may be empty)
        """
        for attempt in range(max_retries):
            available_pods = self.list_available_pods()

            if available_pods:
                logger.debug(
                    f"Found {len(available_pods)} available pod(s) on attempt {attempt + 1}"
                )
                return available_pods

            if attempt < max_retries - 1:
                logger.debug(
                    f"No running pods found on attempt {attempt + 1}, "
                    f"waiting {retry_delay}s before retry"
                )
                time.sleep(retry_delay)
            else:
                logger.debug(f"No running pods found after {max_retries} attempts")

        return []

    def _filter_ready_pods(self, pods) -> List[str]:
        """
        Filter pods to return only ready ones.

        Args:
            pods: Kubernetes pod list object

        Returns:
            List[str]: Names of ready pods
        """
        available = []
        for pod in pods.items:
            if self._is_pod_ready(pod):
                available.append(pod.metadata.name)

        logger.debug(f"Found {len(available)} running pods in namespace {self.namespace}")
        return available

    def _is_pod_ready(self, pod) -> bool:
        """
        Check if a pod is ready (Running phase with ready containers).

        Args:
            pod: Kubernetes pod object

        Returns:
            bool: True if pod is ready, False otherwise
        """
        pod_name = pod.metadata.name

        if pod.status.phase != "Running":
            logger.debug(f"Skipping pod {pod_name}: phase={pod.status.phase}")
            return False

        return self._check_containers_ready(pod, pod_name)

    def _check_containers_ready(self, pod, pod_name: str) -> bool:
        """
        Check if all containers in a pod are ready.

        Args:
            pod: Kubernetes pod object
            pod_name: Name of the pod

        Returns:
            bool: True if all containers are ready
        """
        if not pod.status.container_statuses:
            logger.debug(f"Pod {pod_name} has no container_statuses yet, trusting Running phase")
            return True

        all_ready = all(container.ready for container in pod.status.container_statuses)
        if not all_ready:
            ready_count = sum(1 for c in pod.status.container_statuses if c.ready)
            logger.debug(
                f"Skipping pod {pod_name}: containers not ready "
                f"({ready_count}/{len(pod.status.container_statuses)})"
            )
            return False

        return True

    @staticmethod
    def _is_connection_error(error_msg: str) -> bool:
        """
        Check if error message indicates a connection pool corruption.

        Common patterns:
        - WebSocketBadStatusException with "Handshake status 200 OK"
        - ApiException with "websocket" in message
        - ApiException(0) with no specific reason

        Args:
            error_msg: The error message string

        Returns:
            bool: True if this is a connection error
        """
        error_lower = error_msg.lower()
        return (
            "handshake" in error_lower
            or "websocket" in error_lower
            or "apiexception: (0)" in error_lower
        )



from typing import Any, Dict, Optional

from kubernetes import client
from langchain_core.tools import ToolException


class KubernetesClient:
    """Client for Kubernetes API."""

    def __init__(self, url: str, token: str, verify_ssl: bool = False):
        """
        Initialize the Kubernetes client.

        Args:
            url: Kubernetes API server URL
            token: Bearer token for authentication
            verify_ssl: Whether to verify SSL certificates
        """
        self.url = url
        self.token = token
        self.verify_ssl = verify_ssl

        self.configuration = client.Configuration()
        self.configuration.host = self.url
        self.configuration.verify_ssl = self.verify_ssl
        self.configuration.api_key = {"authorization": f"Bearer {self.token}"}

        self.__api_client: Optional[client.ApiClient] = None

    @property
    def _client(self) -> client.ApiClient:
        """Get or create Kubernetes API client (lazy initialization)."""
        if self.__api_client is None:
            try:
                self.__api_client = client.ApiClient(self.configuration)
            except Exception as e:
                raise ToolException(f"Failed to initialize Kubernetes client: {str(e)}")
        return self.__api_client

    @_client.setter
    def _client(self, value: client.ApiClient) -> None:
        """Set the Kubernetes API client (useful for testing)."""
        self.__api_client = value

    def call_api(
        self,
        suburl: str,
        method: str,
        body: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> str:
        """
        Make an API call to Kubernetes.

        Args:
            suburl: Relative API path (must start with /)
            method: HTTP method (GET, POST, PUT, DELETE, PATCH)
            body: Optional request body
            headers: Optional request headers

        Returns:
            str: Response data as string

        Raises:
            ToolException: If API call fails
        """
        try:
            # Validate suburl
            if not suburl.startswith("/"):
                raise ToolException(f"Kubernetes API path must start with '/', got: {suburl}")

            # Prepare call_api parameters
            call_params = {
                "auth_settings": ["BearerToken"],
                "response_type": "json",
                "_preload_content": False,
            }

            if headers:
                call_params["header_params"] = headers

            if body:
                call_params["body"] = body

            # Make the API call
            response = self._client.call_api(suburl, method.upper(), **call_params)

            # Parse response
            # response is a tuple: (response_data, status_code, response_headers)
            response_data = response[0]
            status_code = response[1]

            # Check status code
            if status_code >= 400:
                error_text = response_data.data.decode("utf-8") if response_data.data else ""
                raise ToolException(f"Kubernetes API error: HTTP {status_code} - {error_text}")

            # Return decoded response data
            if response_data and response_data.data:
                return response_data.data.decode("utf-8")
            else:
                return "Success: Operation completed successfully"

        except ToolException:
            raise
        except Exception as e:
            raise ToolException(f"Kubernetes API call failed: {str(e)}")

    def health_check(self):
        """
        Check if Kubernetes cluster is accessible by querying /version endpoint.

        Raises:
            ToolException: If cluster is not accessible
        """
        self.call_api("/version", "GET")



from typing import List

from codemie_tools.base.base_toolkit import DiscoverableToolkit
from codemie_tools.base.models import ToolKit, ToolSet, Tool
from codemie_tools.core.project_management.confluence.tools import GenericConfluenceTool
from codemie_tools.core.project_management.confluence.tools_vars import GENERIC_CONFLUENCE_TOOL
from codemie_tools.core.project_management.jira.tools import GenericJiraIssueTool
from codemie_tools.core.project_management.jira.tools_vars import GENERIC_JIRA_TOOL


class ProjectManagementToolkitUI(ToolKit):
    toolkit: ToolSet = ToolSet.PROJECT_MANAGEMENT
    tools: List[Tool] = [
        Tool.from_metadata(GENERIC_CONFLUENCE_TOOL, tool_class=GenericConfluenceTool),
        Tool.from_metadata(GENERIC_JIRA_TOOL, tool_class=GenericJiraIssueTool),
    ]
    label: str = ToolSet.PROJECT_MANAGEMENT.value


class ProjectManagementToolkit(DiscoverableToolkit):

    @classmethod
    def get_definition(cls):
        return ProjectManagementToolkitUI()



"""
Local code execution using embedded kernel.

This module provides local code execution capabilities using an embedded kernel.
It's designed for development environments and scenarios where full isolation is not required.

WARNING: Local execution has limited security controls compared to sandbox mode.
Use sandbox mode for production or multi-tenant environments.
"""
import asyncio
import io
import json
import logging
from typing import Any

from jupyter_client.asynchronous.client import AsyncKernelClient
from jupyter_client.manager import AsyncKernelManager
from pydantic import BaseModel
from tenacity import retry, retry_if_exception_cause_type, stop_after_attempt, wait_fixed

logger = logging.getLogger(__name__)


class RuntimeOutput(BaseModel):
    """Output from code execution."""
    type: str
    content: str


class RuntimeOutputTraceback(BaseModel):
    """Output with traceback information."""
    type: str
    content: str
    traceback: str | None = None


UNINITIALIZED_CLIENT_ERROR = RuntimeOutput(
    type="error", content="attempted to use an uninitialized client"
)

_TEXT_PLAIN_MIME = "text/plain"
_IMAGE_PNG_MIME = "image/png"


class LocalKernelExecutor:
    """
    Local kernel executor for running Python code in an embedded kernel.

    This executor provides lightweight code execution without full isolation.
    It's suitable for development and testing but should not be used in
    production or multi-tenant environments.
    """

    def __init__(self, cache_dir: str, is_tcp: bool = True) -> None:
        """
        Initialize the local kernel executor.

        Args:
            cache_dir: Directory for kernel connection files
            is_tcp: Use TCP connection (True) or IPC (False)
        """
        if is_tcp:
            self.connection_config = {
                "shell_port": 0,
                "iopub_port": 0,
                "stdin_port": 0,
                "control_port": 0,
                "ip": "0.0.0.0",
                "transport": "tcp",
                "kernel_name": "python3",
                "hb": "tcp://0.0.0.0:0",
            }
        else:
            self.connection_config = {
                "shell_port": 0,
                "iopub_port": 0,
                "stdin_port": 0,
                "control_port": 0,
                "ip": "localhost",
                "transport": "ipc",
                "kernel_name": "python3",
                "shell": f"ipc://{cache_dir}/shell",
                "iopub": f"ipc://{cache_dir}/iopub",
                "stdin": f"ipc://{cache_dir}/stdin",
                "control": f"ipc://{cache_dir}/control",
                "hb": f"ipc://{cache_dir}/hb",
            }

        self.connection_file = f"{cache_dir}/kernel-connection.json"

        with open(self.connection_file, "w") as f:
            json.dump(self.connection_config, f)

        self.km: AsyncKernelManager | None = None
        self.kc: AsyncKernelClient | None = None

    def _is_relevant_message(self, msg: dict, msg_id: str) -> bool:
        """Check if message is relevant to the current execution."""
        parent_id = msg.get("parent_header", {}).get("msg_id")
        return parent_id == msg_id

    def _handle_stream(self, msg: dict, result_buffer: io.StringIO) -> RuntimeOutput | None:
        """Handle stream messages."""
        text = msg.get("content", {}).get("text", "").strip()
        if text and "Requirement already satisfied:" not in text:
            result_buffer.write(text + "\n")
        return None

    def _handle_execute_result(self, msg: dict, result_buffer: io.StringIO) -> RuntimeOutput | None:
        """Handle execute result messages."""
        data = msg.get("content", {}).get("data", {})
        text = data.get(_TEXT_PLAIN_MIME, "").strip()
        if text:
            result_buffer.write(text + "\n")
        return None

    def _handle_status(self, msg: dict, result_buffer: io.StringIO) -> RuntimeOutput | None:
        """Handle status messages."""
        if msg.get("content", {}).get("execution_state") == "idle":
            # Limit result buffer size
            current_result = result_buffer.getvalue()
            if len(current_result) > 500:
                trimmed_result = "[...]\n" + current_result[-500:]
                return RuntimeOutput(type="text", content=trimmed_result)

            return RuntimeOutput(
                type="text", content=current_result or "code run successfully (no output)"
            )

        return None

    def _handle_error(self, msg: dict, result_buffer: io.StringIO) -> RuntimeOutput | None:
        """Handle error messages."""
        content = msg.get("content", {})
        ename = content.get("ename", "Unknown error")
        evalue = content.get("evalue", "No details")
        error = f"{ename}: {evalue}"
        return RuntimeOutput(type="error", content=error)

    def _handle_display_data(self, msg: dict) -> tuple[str | None, RuntimeOutput | None]:
        """Handle display data messages."""
        data = msg.get("content", {}).get("data", {})

        if _IMAGE_PNG_MIME in data:
            return None, RuntimeOutput(type=_IMAGE_PNG_MIME, content=data[_IMAGE_PNG_MIME])

        if _TEXT_PLAIN_MIME in data:
            return None, RuntimeOutput(type="text", content=data[_TEXT_PLAIN_MIME])

        return None, RuntimeOutput(type="error", content="Could not parse output")

    async def arun(self, code: str) -> RuntimeOutput:
        """
        Execute code asynchronously and return the result.

        Automatically sets up matplotlib to use inline backend for plot capture.
        When plt.show() is called, plots are automatically captured as PNG images.

        Args:
            code: Python code to execute

        Returns:
            RuntimeOutput with execution result
        """
        if not self.kc:
            return UNINITIALIZED_CLIENT_ERROR

        # Set up matplotlib inline backend for automatic plot capture
        # This allows plt.show() to work without plt.savefig()
        setup_code = """
import matplotlib
matplotlib.use('module://matplotlib_inline.backend_inline')
import matplotlib.pyplot as plt
from IPython.display import set_matplotlib_formats
set_matplotlib_formats('png')
"""
        # Execute setup code first (don't wait for response)
        self.kc.execute(setup_code, silent=True)

        # Execute the actual user code
        msg_id = self.kc.execute(code)
        result = io.StringIO()

        while True:
            received_msg = await self.kc.get_iopub_msg()

            msg_type = received_msg.get("header", {}).get("msg_type")

            if msg_type == "display_data":
                _, output = self._handle_display_data(received_msg)
                if output:
                    return output
                continue

            if not self._is_relevant_message(received_msg, msg_id):
                continue

            output = self._handle_message_type(received_msg, msg_type, result)
            if output:
                return output

    def _handle_message_type(self, received_msg: dict, msg_type: Any, result: io.StringIO):
        """Dispatch message to appropriate handler."""
        handler_map = {
            "stream": self._handle_stream,
            "execute_result": self._handle_execute_result,
            "status": self._handle_status,
            "error": self._handle_error,
        }

        handler = handler_map.get(msg_type)

        if handler:
            output = handler(received_msg, result)

            if output:
                return output

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_fixed(1),
        retry=retry_if_exception_cause_type(Exception),
    )
    async def __aenter__(self):
        """Start the kernel manager and client."""
        km: AsyncKernelManager = AsyncKernelManager(connection_file=self.connection_file)
        self.km = km
        await km.start_kernel()

        self.kc = km.client()
        self.kc.start_channels()
        await self.kc.wait_for_ready()

        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Shutdown the kernel manager and client."""
        try:
            if self.kc:
                self.kc.shutdown()
                self.kc.stop_channels()

            if self.km:
                await self.km.shutdown_kernel()
        finally:
            self.km = None
            self.kc = None



import ast
import json
import logging
import re
from typing import Type, Union, Dict, Any

import tiktoken
from langchain_core.tools import ToolException
from pydantic import BaseModel, ValidationError

logger = logging.getLogger(__name__)

def sanitize_string(input_string: str) -> str:
    """
    Sanitize a string by replacing or masking potentially sensitive information.

    This function uses predefined regular expressions to identify and replace common patterns
    of sensitive data such as passwords, usernames, IP addresses, email addresses,
    API keys and credit card numbers.

    Args:
        input_string (str): The original string to be sanitized.

    Returns:
        str: The sanitized string with sensitive information removed or masked.

    Example:
        >>> original_string = "Error: Unable to connect. Username: admin, Password: secret123, IP: 192.168.1.1"
        >>> sanitize_string(original_string)
        'Error: Unable to connect. Username: ***, Password: ***, IP: [IP_ADDRESS]'
    """
    patterns = [
        (r'\b(password|pwd|pass)(\s*[:=]\s*|\s+)(\S+)', r'\1\2***'),  # Passwords
        (r'\b(username|user|uname)(\s*[:=]\s*|\s+)(\S+)', r'\1\2***'),  # Usernames
        (r'\b(?:\d{1,3}\.){3}\d{1,3}\b', '[IP_ADDRESS]'),  # IP addresses
        (r'\b(?:[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,})\b', '[EMAIL]'),  # Email addresses
        (r'\b(api[_-]?key|access[_-]?token)(\s*[:=]\s*|\s+)(\S+)', r'\1\2[API_KEY]'),  # API keys and access tokens
        (r'([?&])(api[_-]?key|key|access[_-]?token)=([A-Za-z0-9_\-]+)', r'\1\2=[API_KEY]'),  # API keys in URL query parameters
        (r'\b(?:\d{4}[-\s]?){4}\b', '[CREDIT_CARD]'),  # Credit card numbers
    ]

    sanitized_string = input_string

    for pattern, replacement in patterns:
        sanitized_string = re.sub(pattern, replacement, sanitized_string, flags=re.IGNORECASE)

    return sanitized_string


def parse_to_dict(input_string):
    """
    Parse a string representation of a dictionary into an actual dictionary.
    Handles both JSON format and Python dictionary syntax.
    """
    if not input_string or not isinstance(input_string, str):
        return {}

    # Remove whitespace
    input_string = input_string.strip()

    # Try parsing as JSON directly
    try:
        return json.loads(input_string)
    except json.JSONDecodeError:
        pass

    # Try parsing as Python literal (for dictionary strings like "{'key': 'value'}")
    try:
        return ast.literal_eval(input_string)
    except (ValueError, SyntaxError):
        pass

    # If the above methods fail, try a more robust approach for Python dict syntax
    if input_string.startswith('{') and input_string.endswith('}'):
        try:
            # Convert Python dict string to proper JSON by handling single quotes
            # This regex replaces single quotes with double quotes while preserving quotes in values
            cleaned = re.sub(r"(?<!\\)'([^']*)'(?!\\)", r'"\1"', input_string)
            # Replace any remaining single quotes with double quotes
            cleaned = cleaned.replace("'", '"')
            return json.loads(cleaned)
        except json.JSONDecodeError:
            # Try a simpler approach as a fallback
            try:
                simple_cleaned = input_string.replace("'", '"')
                return json.loads(simple_cleaned)
            except json.JSONDecodeError:
                pass

    # If all parsing attempts failed, log the error and return empty dict
    print(f"Failed to parse string to dict: {input_string}")
    return {}

OPEN_AI_TOOL_NAME_LIMIT = 64

def parse_tool_input(args_schema: Type[BaseModel], tool_input: Union[str, Dict]):
    try:
        input_args = args_schema
        logger.info(f"Starting parser with input: {tool_input}")
        if isinstance(tool_input, str):
            logger.info("isinstance(tool_input, str)")
            params = parse_to_dict(tool_input)
            result = input_args.model_validate(dict(params))
            return {
                k: getattr(result, k)
                for k, v in result.model_dump().items()
                if k in tool_input
            }
        else:
            logger.info("else isinstance(tool_input, dict)")
            if input_args is not None:
                result = input_args.model_validate(tool_input)
                return {
                    k: getattr(result, k)
                    for k, v in result.model_dump().items()
                    if k in tool_input
                }
        return tool_input
    except Exception as e:
        raise ToolException(f"""
                Cannot parse input parameters.
                Got wrong input: {tool_input}. See description of input parameters.
                Error: {e}
                """)


def clean_json_string(json_string):
    """
    Extract JSON object from a string, removing extra characters before '{' and after '}'.
    Handles both properly formatted JSON and compressed JSON with backslash line continuations.

    This function intelligently handles two cases:
    1. Properly formatted JSON with structural newlines (formatting whitespace)
    2. Compressed JSON with unescaped newlines inside string values

    Args:
        json_string (str): Input string containing a JSON object.

    Returns:
        str: Cleaned JSON string or original string if no JSON object found.
    """
    # Step 1: Remove backslash line continuations (\ followed by newline and optional spaces)
    # This handles Python/shell-style line continuations that aren't valid in JSON
    json_string = re.sub(r"\\\s*\n\s*", "", json_string)

    # Step 2: Extract JSON object (content between outermost { })
    pattern = r"^[^{]*({.*})[^}]*$"
    match = re.search(pattern, json_string, re.DOTALL)
    if match:
        json_string = match.group(1)

    # Step 3: Try parsing as-is first (for properly formatted JSON)
    # If it's valid JSON, return it unchanged to preserve formatting
    try:
        json.loads(json_string)
        return json_string  # Valid JSON, return as-is
    except json.JSONDecodeError:
        # Parsing failed - likely has unescaped newlines in string values
        # Escape newlines and carriage returns for compressed JSON
        json_string = json_string.replace("\n", "\\n").replace("\r", "\\r")
        return json_string


def get_encoding(llm_model: str) -> 'tiktoken.core.Encoding':
    try:
        encoding = tiktoken.encoding_for_model(llm_model)
    except Exception:
        logger.debug(f"Cannot find encoding for model {llm_model}. Using o200k_base encoding")
        encoding = tiktoken.get_encoding("o200k_base")
    return encoding


def humanize_error(error: Exception) -> str:
    """
    If an error is a Pyndatic ValidationError, return a human-readable string
    Otherwise, return the string representation of the error.
    """
    if not isinstance(error, ValidationError):
        return str(error)

    try:
      return ", ".join([
        f"{_format_pydantic_validation_loc(item['loc'])}: {item['msg'].lower()}"
        for item in error.errors()
      ]).capitalize()
    except Exception:
        logger.error("Error formatting Pydantic ValidationError", exc_info=True)
        return str(error)


def _format_pydantic_validation_loc(items):
  """Humanize the location field of a Pydantic validation error"""
  return ".".join(str(loc) for loc in items)


def parse_and_escape_args(query: Union[str, Dict[str, Any]], item_type: str = None) -> Dict[str, Any]:
    """
    Parses arguments from string or dictionary input.
    Handles JSON string parsing with proper error messaging.

    Args:
        query: Input query as either a JSON string or dictionary
        item_type: Optional name of the item type for error messages

    Returns:
        Dictionary of parsed arguments or None if query is None

    Raises:
        ValueError: If query is not a valid JSON string or dictionary
        ToolException: If parsing fails due to invalid format
    """
    if not query:
        return {}

    if isinstance(query, dict):
        return query

    if isinstance(query, str):
        try:
            # Remove markdown code block markers if present
            query_str = query.replace("```json", "").replace("```", "").strip()
            return json.loads(query_str)
        except json.JSONDecodeError as e:
            type_context = f" in {item_type}" if item_type else ""
            raise ToolException(f"Invalid JSON format{type_context}: {str(e)}")

    type_context = f"{item_type}" if item_type else "Input"
    raise ToolException(f"{type_context} must be a JSON string or dict, got {type(query)}")



"""GitLab Toolkit."""
import logging
from typing import List, Optional, Dict, Any

from langchain_core.language_models import BaseChatModel
from langchain_core.tools import BaseTool

from codemie_tools.base.base_toolkit import BaseToolkit
from codemie_tools.git.gitlab.gitlab_openai_tools import (
    OpenAIUpdateFileWholeTool, OpenAIUpdateFileDiffTool,
    CreateFileTool, DeleteFileTool, CreatePRTool, ListBranchesTool, SetActiveBranchTool, CreateGitLabBranchTool,
    GetPullRequesChanges, CreatePullRequestChangeComment
)
from codemie_tools.git.utils import GitCredentials, init_gitlab_api_wrapper

logger = logging.getLogger(__name__)


class CustomGitLabToolkit(BaseToolkit):
    git_creds: GitCredentials
    api_wrapper: Optional[Any] = None
    llm_model: Optional[Any] = None

    @classmethod
    def get_tools_ui_info(cls, *args, **kwargs):
        # no need this function at this moment
        pass

    @classmethod
    def get_toolkit(cls,
                    configs: Dict[str, Any],
                    llm_model: Optional[BaseChatModel] = None):
        git_creds = GitCredentials(**configs)
        api_wrapper = init_gitlab_api_wrapper(git_creds)
        return CustomGitLabToolkit(git_creds=git_creds,
                                   api_wrapper=api_wrapper,
                                   llm_model=llm_model)

    def get_tools(self) -> List[BaseTool]:
        tools = [
            CreateFileTool(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            DeleteFileTool(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            CreatePRTool(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            ListBranchesTool(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            SetActiveBranchTool(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            CreateGitLabBranchTool(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            GetPullRequesChanges(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            CreatePullRequestChangeComment(api_wrapper=self.api_wrapper, credentials=self.git_creds),
            OpenAIUpdateFileWholeTool(api_wrapper=self.api_wrapper, credentials=self.git_creds,
                                      llm_model=self.llm_model),
            OpenAIUpdateFileDiffTool(api_wrapper=self.api_wrapper, credentials=self.git_creds,
                                     llm_model=self.llm_model)
        ]
        return tools
